"""Tests of BasisOfFData."""

import numpy as np
import pytest

from skfda.representation.basis import CustomBasis, FDataBasis, FourierBasis
from skfda.representation.grid import FDataGrid


def test_grid():
    """Test a datagrid toy example."""
    grid_points = np.array([0, 1, 2])
    sample = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [4, 5, 6]]),
        grid_points=np.array([[0, 1, 2]]),
    )

    data_basis = FDataBasis(
        basis=CustomBasis(fdata=sample),
        coefficients=np.array([[1, 0], [0, 1], [1, 1]]),
    )

    evaluated = [
        np.array([1, 2, 3]),
        np.array([4, 5, 6]),
        np.array([5, 7, 9]),
    ]

    np.testing.assert_equal(data_basis(grid_points)[..., 0], evaluated)


def test_basis():
    """Test a databasis toy example."""
    basis = FourierBasis(n_basis=3)
    coeficients = np.array(
        [
            [2, 1, 0],
            [3, 1, 0],
            [1, 2, 9],
        ],
    )

    data_basis = FDataBasis(
        basis=CustomBasis(
            fdata=FDataBasis(basis=basis, coefficients=coeficients),
        ),
        coefficients=coeficients,
    )

    combined_basis = FDataBasis(
        basis=basis,
        coefficients=coeficients @ coeficients,
    )

    eval_points = np.linspace(0, 1, 100)
    np.testing.assert_almost_equal(
        data_basis(eval_points),
        combined_basis(eval_points),
    )


def test_not_linearly_independent_too_many():
    """
    Test that a non linearly independent basis raises an error.

    In this case, the number of samples is greater than the number
    of sampling points or base functions in the underlying base.
    """
    sample = FDataGrid(
        data_matrix=np.array(
            [[1, 2, 3], [2, 4, 6], [15, 4, -2], [1, 28, 0]],
        ),
        grid_points=np.array([[0, 1, 2]]),
    )

    with pytest.raises(ValueError):
        CustomBasis(fdata=sample)

    sample = FDataBasis(
        basis=FourierBasis(n_basis=3),
        coefficients=np.array(
            [
                [0, 1, 0],
                [1, 0, 0],
                [0, 0, 1],
                [1, 1, 1],
            ],
        ),
    )

    with pytest.raises(ValueError):
        CustomBasis(fdata=sample)


def test_not_linearly_independent_range():
    """
    Test that a non linearly independent basis raises an error.

    In this case, the number of samples is valid but the given
    samples are not linearly independent.
    """
    sample = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [2, 4, 6]]),
        grid_points=np.array([[0, 1, 2]]),
    )

    with pytest.raises(ValueError):
        CustomBasis(fdata=sample)

    sample = FDataBasis(
        basis=FourierBasis(n_basis=3),
        coefficients=np.array(
            [
                [2, 1, 0],
                [3, 1, 0],
                [1, 2, 0],
            ],
        ),
    )

    with pytest.raises(ValueError):
        CustomBasis(fdata=sample)


def test_derivative_grid():
    """Test the derivative of a basis constructed from a FDataGrid."""
    base_functions = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [1, 1, 5]]),
        grid_points=np.array([[0, 1, 2]]),
    )
    # The derivative of the first function is always 1
    # The derivative of the second function is 0 and then 4

    basis = CustomBasis(fdata=base_functions)

    coefs = np.array([[1, 0], [0, 1], [1, 1]])
    derivate_basis, derivative_coefs = basis.derivative_basis_and_coefs(
        coefs=coefs,
        order=1,
    )

    derivative = FDataBasis(
        basis=derivate_basis,
        coefficients=derivative_coefs,
    )

    eval_points = np.array([0.5, 1.5])

    # Derivative of the original functions sampled in the given points
    # and multiplied by the coefs
    derivative_evaluated = np.array([[1, 1], [0, 4], [1, 5]])
    np.testing.assert_allclose(
        derivative(eval_points)[..., 0],
        derivative_evaluated,
        atol=1e-15,
    )


def test_derivative_basis():
    """Test the derivative of a basis constructed from a FDataBasis."""
    basis_coef = np.array(
        [
            [2, 1, 0, 10, 0],
            [3, 1, 99, 15, 99],
            [0, 2, 9, 22, 0],
        ],
    )

    base_functions = FDataBasis(
        basis=FourierBasis(n_basis=5),
        coefficients=basis_coef,
    )

    basis = CustomBasis(fdata=base_functions)

    coefs = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1], [2, 1, 5]])
    derivate_basis, derivative_coefs = basis.derivative_basis_and_coefs(
        coefs=coefs,
        order=1,
    )

    derivative = FDataBasis(
        basis=derivate_basis,
        coefficients=derivative_coefs,
    )

    eval_points = np.linspace(0, 1, 10)

    # Derivative of the original functions
    basis_derivative, coefs_derivative = FourierBasis(
        n_basis=5,
    ).derivative_basis_and_coefs(coefs=coefs @ basis_coef)

    derivative_on_the_basis = FDataBasis(
        basis=basis_derivative,
        coefficients=coefs_derivative,
    )
    np.testing.assert_almost_equal(
        derivative(eval_points),
        derivative_on_the_basis(eval_points),
    )


def test_multivariate_codomain():
    """Test basis from a multivariate function."""
    points = np.array([0, 1, 2])
    base_functions = FDataGrid(
        data_matrix=np.array(
            [
                [[0, 0, 1], [0, 0, 2], [0, 0, 3]],
                [[1, 0, 0], [2, 0, 0], [3, 0, 0]],
            ],
        ),
        grid_points=points,
    )
    base = CustomBasis(fdata=base_functions)

    coefs = np.array([[1, 0], [0, 1], [1, 1]])

    functions = FDataBasis(
        basis=base,
        coefficients=coefs,
    )

    expected_data = np.array(
        [
            [[0, 0, 1], [0, 0, 2], [0, 0, 3]],
            [[1, 0, 0], [2, 0, 0], [3, 0, 0]],
            [[1, 0, 1], [2, 0, 2], [3, 0, 3]],
        ],
    )
    np.testing.assert_equal(functions(points), expected_data)


def test_multivariate_codomain_linearly_dependent():
    """Test basis from multivariate linearly dependent functions."""
    points = np.array([0, 1, 2])
    base_functions = FDataGrid(
        data_matrix=np.array(
            [
                [[0, 0, 1], [0, 0, 2], [0, 0, 3]],
                [[1, 0, 0], [2, 0, 0], [3, 0, 0]],
                [[1, 0, 1], [2, 0, 2], [3, 0, 3]],
            ],
        ),
        grid_points=points,
    )
    # The third function is the sum of the first two

    with pytest.raises(ValueError):
        CustomBasis(fdata=base_functions)


def test_evaluate_derivative():
    """Test the evaluation of the derivative of a DataBasis."""
    grid_points = np.array([[0, 1, 2]])
    base_functions = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [1, 3, 5]]),
        grid_points=grid_points,
    )
    basis = CustomBasis(fdata=base_functions)

    coefs = np.array([[1, 0], [0, 1], [1, 2]])

    functions = FDataBasis(
        basis=basis,
        coefficients=coefs,
    )
    deriv = functions.derivative(order=1)
    derivate_evalated = deriv(np.array([0.5, 1.5]))
    np.testing.assert_allclose(
        derivate_evalated[..., 0],
        np.array([[1, 1], [2, 2], [5, 5]]),
        rtol=1e-15,
    )


def test_coordinates():
    """Test the coordinates of a basis."""
    grid_points = np.array([[0, 1]])
    base_functions = FDataGrid(
        data_matrix=np.array([
            [[1, 2, 3], [1, 3, 5]],
            [[4, 5, 6], [7, 8, 9]],
        ]),
        grid_points=grid_points,
    )
    basis = CustomBasis(fdata=base_functions)

    coefs = np.array([[1, 0], [0, 1], [1, 2]])

    functions = FDataBasis(
        basis=basis,
        coefficients=coefs,
    )

    # First coordinate at 0
    np.testing.assert_allclose(
        functions.coordinates[0](0),
        np.array([[[1]], [[4]], [[9]]]),
    )

    # Second two coordinates at 1
    np.testing.assert_allclose(
        functions.coordinates[1:3](1),
        np.array([[[3, 5]], [[8, 9]], [[16 + 3, 18 + 5]]]),
    )


def test_equality():
    """Test the equality of two basis."""
    grid_points = np.array([[0, 1, 2]])
    base_functions = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [1, 3, 5]]),
        grid_points=grid_points,
    )
    basis = CustomBasis(fdata=base_functions)

    other_base_functions = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [1, 3, 5]]),
        grid_points=grid_points,
    )

    other_basis = CustomBasis(fdata=other_base_functions)

    assert basis == other_basis

    different_base_functions = FDataGrid(
        data_matrix=np.array([[1, 2, 3], [1, 8, 5]]),
        grid_points=grid_points,
    )

    different_basis = CustomBasis(fdata=different_base_functions)

    assert basis != different_basis



# -*- coding: utf-8 -*-

"""Hat Matrix.

This module include implementation to create Nadaraya-Watson,
Local Linear Regression and K-Nearest Neighbours hat matrices used in
kernel smoothing and kernel regression.

"""
from __future__ import annotations

import abc
import math
from typing import Callable, Final, TypeVar, Union, overload

import numpy as np

from .._utils._sklearn_adapter import BaseEstimator
from ..representation._functional_data import FData
from ..representation.basis import FDataBasis
from ..typing._numpy import NDArrayFloat
from . import kernels

Input = TypeVar("Input", bound=Union[FData, NDArrayFloat])
Prediction = TypeVar("Prediction", bound=Union[NDArrayFloat, FData])

DEFAULT_BANDWIDTH_PERCENTILE: Final = 15


class HatMatrix(
    BaseEstimator,
):
    """
    Hat Matrix.

    Base class for different hat matrices.

    See also:
        :class:`~skfda.misc.hat_matrix.NadarayaWatsonHatMatrix`
        :class:`~skfda.misc.hat_matrix.LocalLinearRegressionHatMatrix`
        :class:`~skfda.misc.hat_matrix.KNeighborsHatMatrix`
    """

    def __init__(
        self,
        *,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.normal,
    ):
        self.kernel = kernel

    @overload
    def __call__(
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: Input,
        X: Input,
        y_train: None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> NDArrayFloat:
        pass

    @overload
    def __call__(
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: Input,
        X: Input,
        y_train: Prediction | None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> Prediction:
        pass

    def __call__(
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: Input,
        X: Input,
        y_train: NDArrayFloat | FData | None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> NDArrayFloat | FData:
        r"""
        Calculate the hat matrix or the prediction.

        If y_train is given, the estimation for X is calculated, otherwise,
        hat matrix, :math:`\hat{H]`, is returned. The prediction, :math:`y`,
        can be calculated as :math:`y = \hat{H} \dot y_train`.

        Args:
            delta_x: Matrix of distances between points or
                functions.
            X_train: Training data.
            X: Test samples.
            y_train: Target values for X_train.
            weights: Weights to be applied to the
                resulting matrix columns.

        Returns:
            The prediction if y_train is given, hat matrix otherwise.

        """
        # Obtain the non-normalized matrix
        matrix = self._hat_matrix_function_not_normalized(
            delta_x=delta_x,
        )

        # Adjust weights
        if weights is not None:
            matrix *= weights

        # Set diagonal to zero if requested (for testing purposes only)
        if _cv:
            np.fill_diagonal(matrix, 0)

        # Renormalize weights
        rs = np.sum(matrix, axis=1)
        rs[rs == 0] = 1

        matrix = (matrix.T / rs).T

        if y_train is None:
            return matrix

        return matrix @ y_train

    @abc.abstractmethod
    def _hat_matrix_function_not_normalized(
        self,
        *,
        delta_x: NDArrayFloat,
    ) -> NDArrayFloat:
        pass


class NadarayaWatsonHatMatrix(HatMatrix):
    r"""Nadaraya-Watson method.

    Creates the matrix :math:`\hat{H}`, used in the kernel smoothing and kernel
    regression algorithms, as explained below

    .. math::
        \hat{H}_{i,j} = \frac{K\left(\frac{d(x_j-x_i')}{h}\right)}{\sum_{k=1}^{
        n}K\left(\frac{d(x_k-x_i')}{h}\right)}

    For smoothing, :math:`\{x_1, ..., x_n\}` are the points with known value
    and :math:`\{x_1', ..., x_m'\}` are the points for which it is desired to
    estimate the smoothed value. The distance :math:`d` is the absolute value
    function :footcite:`wasserman_2006_nonparametric`.

    For regression, :math:`\{x_1, ..., x_n\}` is the functional data and
    :math:`\{x_1', ..., x_m'\}` are the functions for which it is desired to
    estimate the scalar value. Here, :math:`d` is some functional distance
    :footcite:`ferraty+vieu_2006_functional`.

    In both cases :math:`K(\cdot)` is a kernel function and :math:`h` is the
    bandwidth.

    Args:
        bandwidth: Window width of the kernel
            (also called h or bandwidth).
        kernel: Kernel function. By default a normal
            kernel.

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        *,
        bandwidth: float | None = None,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.normal,
    ):
        super().__init__(kernel=kernel)
        self.bandwidth = bandwidth

    def _hat_matrix_function_not_normalized(
        self,
        *,
        delta_x: NDArrayFloat,
    ) -> NDArrayFloat:

        bandwidth = (
            float(np.percentile(delta_x, DEFAULT_BANDWIDTH_PERCENTILE))
            if self.bandwidth is None
            else self.bandwidth
        )

        return self.kernel(delta_x / bandwidth)


class LocalLinearRegressionHatMatrix(HatMatrix):
    r"""Local linear regression method.

    Creates the matrix :math:`\hat{H}`, used in the kernel smoothing and kernel
    regression algorithms, as explained below.

    For **kernel smoothing** algorithm to estimate the smoothed value for
    :math:`t_i'` the following error must be minimised

    .. math::
        AWSE(a, b) = \sum_{j=1}^n \left[ \left(y_j -
        \left(a +  b (t_j - t'_i) \right) \right)^2
        K \left( \frac {|t_j - t'_i|}{h} \right) \right ]

    which gives the following expression for each cell

    .. math::
        \hat{H}_{i,j} = \frac{b_j(t_i')}{\sum_{k=1}^{n}b_k(t_i')}

    .. math::
        b_j(t_i') = K\left(\frac{t_j - t_i'}{h}\right) S_{n,2}(t_i') -
        (t_j - t_i')S_{n,1}(t_i')

    .. math::
        S_{n,k}(t_i') = \sum_{j=1}^{n}K\left(\frac{t_j-t_i'}{h}\right)
        (t_j-t_i')^k

    where :math:`\{t_1, t_2, ..., t_n\}` are points with known value and
    :math:`\{t_1', t_2', ..., t_m'\}` are the points for which it is
    desired to estimate the smoothed value
    :footcite:`wasserman_2006_nonparametric`.

    For **kernel regression** algorithm:

    Given functional data, :math:`\{X_1, X_2, ..., X_n\}` where each function
    is expressed in a orthonormal basis with :math:`J` elements and scalar
    response :math:`Y = \{y_1, y_2, ..., y_n\}`.

    It is desired to estimate the values
    :math:`\hat{Y} = \{\hat{y}_1, \hat{y}_2, ..., \hat{y}_m\}`
    for the data :math:`\{X'_1, X'_2, ..., X'_m\}`
    (expressed in the same basis).

    For each :math:`X'_k` the estimation :math:`\hat{y}_k` is obtained by
    taking the value :math:`a_k` from the vector
    :math:`(a_k, b_{1k}, ..., b_{Jk})` which minimizes the following expression

    .. math::
        AWSE(a_k, b_{1k}, ..., b_{Jk}) = \sum_{i=1}^n \left(y_i -
        \left(a_k + \sum_{j=1}^J b_{jk} c_{ik}^j \right) \right)^2
        K \left( \frac {d(X_i - X'_k)}{h} \right)

    Where :math:`c_{ik}^j` is the :math:`j`-th coefficient in a truncated basis
    expansion of :math:`X_i - X'_k = \sum_{j=1}^J c_{ik}^j` and :math:`d` some
    functional distance :footcite:`baillo+grane_2009_local`

    For both cases, :math:`K(\cdot)` is a kernel function and :math:`h` the
    bandwidth.

    Args:
        bandwidth: Window width of the kernel
            (also called h).
        kernel: Kernel function. By default a normal
            kernel.

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        *,
        bandwidth: float | None = None,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.normal,
    ):
        super().__init__(kernel=kernel)
        self.bandwidth = bandwidth

    @overload
    def __call__(
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: FData | NDArrayFloat,
        X: FData | NDArrayFloat,
        y_train: None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> NDArrayFloat:
        pass

    @overload
    def __call__(
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: FData | NDArrayFloat,
        X: FData | NDArrayFloat,
        y_train: Prediction | None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> Prediction:
        pass

    def __call__(  # noqa: D102
        self,
        *,
        delta_x: NDArrayFloat,
        X_train: FData | NDArrayFloat,
        X: FData | NDArrayFloat,
        y_train: NDArrayFloat | FData | None = None,
        weights: NDArrayFloat | None = None,
        _cv: bool = False,
    ) -> NDArrayFloat | FData:

        bandwidth = (
            float(np.percentile(delta_x, DEFAULT_BANDWIDTH_PERCENTILE))
            if self.bandwidth is None
            else self.bandwidth
        )

        # Smoothing for functions of one variable
        if not isinstance(X_train, FDataBasis) and X_train[0].shape[0] == 1:
            delta_x = np.subtract.outer(
                X.flatten(),
                X_train.flatten(),
            )

            assert y_train is None

            return super().__call__(  # noqa: WPS503
                delta_x=delta_x,
                X_train=X_train,
                X=X,
                y_train=y_train,
                weights=weights,
                _cv=_cv,
            )

        # Regression
        if isinstance(X_train, FData):
            assert isinstance(X, FData)

            if not (
                isinstance(X_train, FDataBasis)
                and isinstance(X, FDataBasis)
            ):
                raise ValueError("Only FDataBasis is supported for now.")

            m1 = X_train.coefficients
            m2 = X.coefficients

            if y_train is None:
                y_train = np.identity(X_train.n_samples)

        else:
            m1 = X_train
            m2 = X

            if y_train is None:
                y_train = np.identity(X_train.shape[0])

        # Subtract previous matrices obtaining a 3D matrix
        # The i-th element contains the matrix X_train - X[i]
        C = m1 - m2[:, np.newaxis]

        # Inner product matrix only is applicable in regression
        if isinstance(X_train, FDataBasis):
            inner_product_matrix = X_train.basis.inner_product_matrix()

            # Calculate new coefficients taking into account cross-products
            # if the basis is orthonormal, C would not change
            C = C @ inner_product_matrix  # noqa: WPS350

        # Adding a column of ones in the first position of all matrices
        dims = (C.shape[0], C.shape[1], 1)
        C = np.concatenate((np.ones(dims), C), axis=-1)

        return self._solve_least_squares(
            delta_x=delta_x,
            coefs=C,
            y_train=y_train,
            bandwidth=bandwidth,
        )

    def _solve_least_squares(
        self,
        delta_x: NDArrayFloat,
        coefs: NDArrayFloat,
        y_train: NDArrayFloat,
        *,
        bandwidth: float,
    ) -> NDArrayFloat:

        W = np.sqrt(self.kernel(delta_x / bandwidth))

        # A x = b
        # Where x = (a, b_1, ..., b_J).
        A = (coefs.T * W.T).T
        b = np.einsum('ij, j... -> ij...', W, y_train)

        # For Ax = b calculates x that minimize the square error
        # From https://stackoverflow.com/questions/42534237/broadcasted-lstsq-least-squares  # noqa: E501
        u, s, vT = np.linalg.svd(A, full_matrices=False)

        uTb = np.einsum('ijk, ij...->ik...', u, b)
        uTbs = (uTb.T / s.T).T
        x = np.einsum('ijk,ij...->ik...', vT, uTbs)

        return x[:, 0]  # type: ignore[no-any-return]

    def _hat_matrix_function_not_normalized(
        self,
        *,
        delta_x: NDArrayFloat,
    ) -> NDArrayFloat:

        if self.bandwidth is None:
            percentage = 15
            self.bandwidth = np.percentile(np.abs(delta_x), percentage)

        k = self.kernel(delta_x / self.bandwidth)

        s1 = np.sum(k * delta_x, axis=1, keepdims=True)  # S_n_1
        s2 = np.sum(k * delta_x ** 2, axis=1, keepdims=True)  # S_n_2
        b = (k * (s2 - delta_x * s1))  # b_i(x_j)

        return b  # type: ignore[no-any-return] # noqa: WPS331


class KNeighborsHatMatrix(HatMatrix):
    r"""K-nearest neighbour kernel method.

    Creates the matrix :math:`\hat{H}`, used in the kernel smoothing and kernel
    regression algorithms, as explained below.

    .. math::
        \hat{H}_{i,j} = \frac{K\left(\frac{d(x_j-x_i')}{h_i}\right)}
        {\sum_{k=1}^{n}K\left(\frac{d(x_k-x_i')}{h_{i}}\right)}

    For smoothing, :math:`\{x_1, ..., x_n\}` are the points with known value
    and :math:`\{x_1', ..., x_m'\}` are the points for which it is desired to
    estimate the smoothed value. The distance :math:`d` is the absolute value.

    For regression, :math:`\{x_1, ..., x_n\}` are the functional data and
    :math:`\{x_1', ..., x_m'\}` are the functions for which it is desired to
    estimate the scalar value. Here, :math:`d` is some functional distance.

    In both cases, :math:`K(\cdot)` is a kernel function and
    :math:`h_{i}` is calculated as the distance from :math:`x_i'` to its
    ``n_neighbors``-th nearest neighbor in :math:`\{x_1, ..., x_n\}`
    :footcite:`ferraty+vieu_2006_computational`.

    Used with the uniform kernel, it takes the average of the closest
    ``n_neighbors`` points to a given point.

    Args:
        n_neighbors: Number of nearest neighbours. By
            default it takes the 5% closest elements.
        kernel: Kernel function. By default a uniform
            kernel to perform a 'usual' k nearest neighbours estimation.

    References:
        .. footbibliography::

    """

    def __init__(
        self,
        *,
        n_neighbors: int | None = None,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.uniform,
    ):
        self.n_neighbors = n_neighbors
        self.kernel = kernel

    def _hat_matrix_function_not_normalized(
        self,
        *,
        delta_x: NDArrayFloat,
    ) -> NDArrayFloat:

        input_points_len = delta_x.shape[1]

        n_neighbors = (
            math.floor(
                np.percentile(
                    range(1, input_points_len),
                    5,
                ),
            )
            if self.n_neighbors is None
            else self.n_neighbors
        )

        if n_neighbors <= 0:
            raise ValueError('h must be greater than 0')

        # Tolerance to avoid points landing outside the kernel window due to
        # computation error
        tol = np.finfo(np.float64).eps

        # For each row in the distances matrix, it calculates the furthest
        # point within the k nearest neighbours
        vec = np.sort(
            delta_x,
            axis=1,
        )[:, n_neighbors - 1] + tol

        return self.kernel((delta_x.T / vec).T)



"""Unsupervised learner for implementing neighbor searches."""
from __future__ import annotations

from typing import Any, TypeVar, Union, overload

from typing_extensions import Literal

from ..._utils._neighbors_base import (
    AlgorithmType,
    KNeighborsMixin,
    RadiusNeighborsMixin,
)
from ...misc.metrics import l2_distance
from ...representation import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat

InputBound = Union[NDArrayFloat, FData]
Input = TypeVar("Input", contravariant=True, bound=InputBound)
SelfType = TypeVar("SelfType", bound="NearestNeighbors[Any]")


class NearestNeighbors(
    KNeighborsMixin[Input, None],
    RadiusNeighborsMixin[Input, None],
):
    """
    Unsupervised learner for implementing neighbor searches.

    Parameters:
        n_neighbors: Number of neighbors to use by default for
            :meth:`kneighbors` queries.
        radius: Range of parameter space to use by default for
            :meth:`radius_neighbors` queries.
        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to BallTree or KDTree.  This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree.  The optimal value depends on the
            nature of the problem.
        metric: The distance metric to use for the tree.  The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
            Doesn't affect :meth:`fit` method.

    Examples:
        Firstly, we will create a toy dataset with 2 classes

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)

        We will fit a Nearest Neighbors estimator

        >>> from skfda.ml.clustering import NearestNeighbors
        >>> neigh = NearestNeighbors(radius=.3)
        >>> neigh.fit(fd)
        NearestNeighbors(...radius=0.3...)

        Now we can query the k-nearest neighbors.

        >>> distances, index = neigh.kneighbors(fd[:2])
        >>> index # Index of k-neighbors of samples 0 and 1
        array([[ 0,  7,  6, 11,  2],...)

        >>> distances.round(2) # Distances to k-neighbors
        array([[ 0.  ,  0.28,  0.29,  0.29,  0.3 ],
               [ 0.  ,  0.27,  0.28,  0.29,  0.3 ]])

        We can query the neighbors in a given radius too.

        >>> distances, index = neigh.radius_neighbors(fd[:2])
        >>> index[0]
        array([ 0,  2,  6,  7, 11]...)

        >>> distances[0].round(2) # Distances to neighbors of the sample 0
        array([ 0.  ,  0.3 ,  0.29,  0.28,  0.29])

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroids`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`


    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.KNeighborsClassifier`.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    @overload
    def __init__(
        self: NearestNeighbors[NDArrayFloat],
        *,
        n_neighbors: int = 5,
        radius: float = 1.0,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: NearestNeighbors[InputBound],
        *,
        n_neighbors: int = 5,
        radius: float = 1.0,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_neighbors: int = 5,
        radius: float = 1.0,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    # Parameters are important
    def __init__(  # noqa: WPS612
        self,
        *,
        n_neighbors: int = 5,
        radius: float = 1.0,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        super().__init__(
            n_neighbors=n_neighbors,
            radius=radius,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

    # There is actually a change here: the default parameter!!
    def fit(  # noqa: WPS612, D102
        self: SelfType,
        X: Input,
        y: None = None,
    ) -> SelfType:
        return super().fit(X, y)



"""Tests for boxplot functionality."""
import unittest

import numpy as np

from skfda import FDataGrid
from skfda.exploratory.depth import IntegratedDepth
from skfda.exploratory.visualization import Boxplot


class TestBoxplot(unittest.TestCase):
    """Test Boxplot."""

    def test_univariate(self) -> None:
        """Test univariate case."""
        data_matrix = [
            [1, 1, 2, 3, 2.5, 2],
            [0.5, 0.5, 1, 2, 1.5, 1],
            [-1, -1, -0.5, 1, 1, 0.5],  # noqa: WPS204
            [-0.5, -0.5, -0.5, -1, -1, -1],
        ]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        fdataBoxplot = Boxplot(fd, depth_method=IntegratedDepth())
        np.testing.assert_array_equal(
            fdataBoxplot.median.ravel(),
            np.array([-1, -1, -0.5, 1, 1, 0.5]),
        )
        np.testing.assert_array_equal(
            fdataBoxplot.central_envelope[0].ravel(),
            np.array([-1, -1, -0.5, -1, -1, -1]),
        )
        np.testing.assert_array_equal(
            fdataBoxplot.central_envelope[1].ravel(),
            np.array([-0.5, -0.5, -0.5, 1, 1, 0.5]),
        )
        np.testing.assert_array_equal(
            fdataBoxplot.non_outlying_envelope[0].ravel(),
            np.array([-1, -1, -0.5, -1, -1, -1]),
        )
        np.testing.assert_array_equal(
            fdataBoxplot.non_outlying_envelope[1].ravel(),
            np.array([-0.5, -0.5, -0.5, 1, 1, 0.5]),
        )
        self.assertEqual(len(fdataBoxplot.envelopes), 1)
        np.testing.assert_array_equal(
            fdataBoxplot.envelopes[0],
            fdataBoxplot.central_envelope,
        )
        np.testing.assert_array_equal(
            fdataBoxplot.outliers,
            np.array([True, True, False, False]),
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Any, Tuple, TypeVar

import multimethod
import numpy as np

from ...typing._numpy import NDArrayFloat
from .._functional_data import FData
from ..grid import FDataGrid
from ._basis import Basis
from ._fdatabasis import FDataBasis

T = TypeVar("T", bound="CustomBasis")


class CustomBasis(Basis):
    """Basis composed of custom functions.

    Defines a basis composed of the functions in the :class:`FData` object
    passed as argument.
    The functions must be linearly independent, otherwise
    an exception is raised.

    Parameters:
        fdata: Functions that define the basis.

    """

    def __init__(
        self,
        *,
        fdata: FData,
    ) -> None:
        """Basis constructor."""
        super().__init__(
            domain_range=fdata.domain_range,
            n_basis=fdata.n_samples,
        )
        self.fdata = fdata
        self._check_linearly_independent(fdata)

    @multimethod.multidispatch
    def _check_linearly_independent(self, fdata: FData) -> None:
        raise ValueError("Unexpected type of functional data object.")

    @_check_linearly_independent.register
    def _check_linearly_independent_grid(self, fdata: FDataGrid) -> None:
        # Reshape to a bidimensional matrix. This only affects FDataGrids
        # whose codomain is not 1-dimensional and it can be done because
        # checking linear independence in (R^n)^k is equivalent to doing
        # it in R^(nk).
        coord_matrix = fdata.data_matrix.reshape(
            fdata.data_matrix.shape[0],
            -1,
        )
        return self._check_linearly_independent_matrix(coord_matrix)

    @_check_linearly_independent.register
    def _check_linearly_independent_basis(self, fdata: FDataBasis) -> None:
        return self._check_linearly_independent_matrix(fdata.coefficients)

    def _check_linearly_independent_matrix(self, matrix: NDArrayFloat) -> None:
        """Check if the functions are linearly independent."""
        if matrix.shape[0] > matrix.shape[1]:
            raise ValueError(
                "There are more functions than the maximum dimension of the "
                "space that they could generate.",
            )

        rank = np.linalg.matrix_rank(matrix)
        if rank != matrix.shape[0]:
            raise ValueError(
                "There are only {rank} linearly independent "
                "functions".format(
                    rank=rank,
                ),
            )

    def _derivative_basis_and_coefs(
        self: T,
        coefs: NDArrayFloat,
        order: int = 1,
    ) -> Tuple[T, NDArrayFloat]:
        deriv_fdata = self.fdata.derivative(order=order)

        return self._create_subspace_basis_coef(deriv_fdata, coefs)

    @multimethod.multidispatch
    def _create_subspace_basis_coef(
        self: T,
        fdata: FData,
        coefs: np.ndarray,
    ) -> Tuple[T, NDArrayFloat]:
        """
        Create a basis of the subspace generated by the given functions.

        Args:
            fdata: The resulting basis will span the subspace generated
                by these functions.
            coefs: Coefficients of some functions in the given fdata.
                These coefficients will be transformed into the coefficients
                of the same functions in the resulting basis.
        """
        raise ValueError(
            "Unexpected type of functional data object: {type}.".format(
                type=type(fdata),
            ),
        )

    @_create_subspace_basis_coef.register
    def _create_subspace_basis_coef_grid(
        self: T,
        fdata: FDataGrid,
        coefs: np.ndarray,
    ) -> Tuple[T, NDArrayFloat]:

        # Reshape to a bidimensional matrix. This can be done because
        # working in (R^n)^k is equivalent to working in R^(nk) when
        # it comes to linear independence and basis.
        data_matrix_reshaped = fdata.data_matrix.reshape(
            fdata.data_matrix.shape[0],
            -1,
        )
        # If the basis formed by the derivatives has maximum rank,
        # we can just return that
        rank = np.linalg.matrix_rank(data_matrix_reshaped)
        if rank == fdata.n_samples:
            return type(self)(fdata=fdata), coefs

        # Otherwise, we need to find the basis of the subspace generated
        # by the functions
        q, r = np.linalg.qr(data_matrix_reshaped.T)

        # Go back from R^(nk) to (R^n)^k
        fdata.data_matrix = q.T.reshape(
            -1,
            *fdata.data_matrix.shape[1:],
        )

        new_basis = type(self)(fdata=fdata)

        # Since the QR decomponsition yields an orthonormal basis,
        # the coefficients are just the projections of values of
        # the functions in every point (coefs @ data_matrix_reshaped)
        # in the new basis (q).
        # Note that to simply the calculations, we use both the data_matrix
        # and the basis matrix in R^(nk) instead of the original space
        values_in_eval_points = coefs @ data_matrix_reshaped
        coefs = values_in_eval_points @ q

        return new_basis, coefs

    @_create_subspace_basis_coef.register
    def _create_subspace_basis_coef_basis(
        self: T,
        fdata: FDataBasis,
        coefs: np.ndarray,
    ) -> Tuple[T, NDArrayFloat]:

        # If the basis formed by the derivatives has maximum rank,
        # we can just return that
        rank = np.linalg.matrix_rank(fdata.coefficients)
        if rank == fdata.n_samples:
            return type(self)(fdata=fdata), coefs

        q, r = np.linalg.qr(fdata.coefficients.T)

        fdata.coefficients = q.T

        new_basis = type(self)(fdata=fdata)

        # Since the QR decomponsition yields an orthonormal basis,
        # the coefficients are just the result of projecting the
        # coefficients in the underlying basis of the FDataBasis onto
        # the new basis (q)
        coefs_wrt_underlying_fdata_basis = coefs @ fdata.coefficients
        coefs = coefs_wrt_underlying_fdata_basis @ q

        return new_basis, coefs

    def _coordinate_nonfull(
        self,
        coefs: NDArrayFloat,
        key: int | slice,
    ) -> Tuple[Basis, NDArrayFloat]:
        return CustomBasis(fdata=self.fdata.coordinates[key]), coefs

    def _evaluate(
        self,
        eval_points: NDArrayFloat,
    ) -> NDArrayFloat:
        return self.fdata(eval_points)

    def _gram_matrix(self) -> NDArrayFloat:
        """
        Compute the Gram matrix.

        Subclasses may override this method for improving computation
        of the Gram matrix.

        """
        from ...misc import inner_product_matrix
        return inner_product_matrix(self.fdata)

    def __len__(self) -> int:
        return self.n_basis

    @property
    def dim_codomain(self) -> int:
        return self.fdata.dim_codomain

    def __eq__(self, other: Any) -> bool:
        return super().__eq__(other) and all(self.fdata == other.fdata)

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        """Representation of a CustomBasis object."""
        return "{super}, fdata={fdata}".format(
            super=super().__repr__(),
            fdata=self.fdata,
        )

    def __hash__(self) -> int:
        return hash(self.fdata)



from __future__ import annotations

from typing import Sequence, TypeVar, Union

import numpy as np
from scipy.linalg import logm
from sklearn.base import clone
from sklearn.utils.validation import check_is_fitted

from ..._utils import _classifier_get_classes
from ..._utils._sklearn_adapter import BaseEstimator, ClassifierMixin
from ...exploratory.stats.covariance import CovarianceEstimator
from ...representation import FDataGrid
from ...typing._numpy import NDArrayFloat, NDArrayInt, NDArrayStr

Target = TypeVar("Target", bound=Union[NDArrayInt, NDArrayStr])


class QuadraticDiscriminantAnalysis(
    BaseEstimator,
    ClassifierMixin[FDataGrid, Target],
):
    """
    Functional quadratic discriminant analysis.

    It is based on the assumption that the data is part of a Gaussian process
    and depending on the output label, the covariance and mean parameters are
    different for each class. This means that curves classified with one
    determined label come from a distinct Gaussian process compared with data
    that is classified with a different label.

    The training phase of the classifier will try to approximate the two
    main parameters of a Gaussian process for each class. The covariance
    will be estimated by fitting the initial kernel passed on the creation of
    the ParameterizedFunctionalQDA object.
    The result of the training function will be two arrays, one of means and
    another one of covariances. Both with length ``n_classes``.

    The prediction phase instead uses a quadratic discriminant classifier to
    predict which gaussian process of the fitted ones correspond the most with
    each curve passed.

    Warning:
        This classifier is experimental as it does not come from a
        peer-published paper.

    Parameters:
        kernel: Initial kernel to be fitted with the training data. For now,
            only kernels that belongs to the GPy module are allowed.

        regularizer: Parameter that regularizes the covariance matrices in
            order to avoid Singular matrices. It is multiplied by the identity
            matrix and then added to the covariance one.


    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> X, y = fetch_growth(return_X_y=True, as_frame=True)
        >>> X = X.iloc[:, 0].values
        >>> y = y.values.codes
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     X,
        ...     y,
        ...     test_size=0.3,
        ...     stratify=y,
        ...     random_state=0,
        ... )

        Then we need to choose and import a kernel so it can be fitted with
        the data in the training phase. We will use a Gaussian kernel. The
        variance and lengthscale parameters will be optimized during the
        training phase. Therefore, the initial values do not matter too much.
        We will use random values such as 1 for the mean and 6 for the
        variance.

        >>> from skfda.exploratory.stats.covariance import (
        ...     ParametricGaussianCovariance
        ... )
        >>> from skfda.ml.classification import QuadraticDiscriminantAnalysis
        >>> from skfda.misc.covariances import Gaussian
        >>> rbf = Gaussian(variance=6, length_scale=1)

        We will fit the ParameterizedFunctionalQDA with training data. We
        use as regularizer parameter a low value such as 0.05.

        >>> qda = QuadraticDiscriminantAnalysis(
        ...     ParametricGaussianCovariance(rbf),
        ...     regularizer=0.05,
        ... )
        >>> qda = qda.fit(X_train, y_train)


        We can predict the class of new samples.

        >>> list(qda.predict(X_test))
        [0, 1, 0, 0, 1, 1, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1,
         1, 0, 1, 0, 1, 0, 1, 1]

        Finally, we calculate the mean accuracy for the test data.

        >>> round(qda.score(X_test, y_test), 2)
        0.96
    """

    means_: Sequence[FDataGrid]

    def __init__(
        self,
        cov_estimator: CovarianceEstimator[FDataGrid],
        *,
        regularizer: float = 0,
    ) -> None:
        self.cov_estimator = cov_estimator
        self.regularizer = regularizer

    def fit(
        self,
        X: FDataGrid,
        y: Target,
    ) -> QuadraticDiscriminantAnalysis[Target]:
        """
        Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape (n_samples).

        Returns:
            self
        """
        classes, y_ind = _classifier_get_classes(y)
        self.classes_ = classes
        self.y_ind = y_ind

        self._fit_gaussian_process(X)

        self.priors_ = self._calculate_priors(y)
        self._log_priors = np.log(self.priors_)

        self._regularized_covariances = (
            self._covariances
            + self.regularizer * np.eye(len(X.grid_points[0]))
        )

        self._log_determinant_covariances = np.asarray([
            np.trace(logm(regularized_covariance))
            for regularized_covariance in self._regularized_covariances
        ])

        return self

    def predict(self, X: FDataGrid) -> Target:
        """
        Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
            for each data sample.
        """
        check_is_fitted(self)

        return self.classes_[  # type: ignore[no-any-return]
            np.argmax(
                self._calculate_log_likelihood(X.data_matrix),
                axis=1,
            )]

    def _calculate_priors(self, y: Target) -> NDArrayFloat:
        """
        Calculate the prior probability of each class.

        Args:
            y: ndarray with the labels of the training data.

        Returns:
            Numpy array with the respective prior of each class.
        """
        _, counts = np.unique(y, return_counts=True)
        return counts / len(y)

    def _fit_gaussian_process(
        self,
        X: FDataGrid,
    ) -> None:
        """
        Fit the kernel to the data in each class.

        For each class the initial kernel passed as parameter is
        adjusted and the mean is calculated.

        Args:
            X: FDataGrid with the training data.

        """
        cov_estimators = []
        means = []
        covariance = []
        for class_index, _ in enumerate(self.classes_):
            X_class = X[self.y_ind == class_index]
            cov_estimator = clone(self.cov_estimator).fit(X_class)

            cov_estimators.append(cov_estimator)
            means.append(cov_estimator.location_)
            # TODO: QDA should use the covariance estimators interface
            covariance.append(
                cov_estimator.covariance_.cov_fdata.data_matrix[0, ..., 0],
            )

        self.means_ = means
        self._covariances = np.asarray(covariance)

    def _calculate_log_likelihood(self, X: NDArrayFloat) -> NDArrayFloat:
        """
        Calculate the log likelihood quadratic discriminant analysis.

        Args:
            X: sample where we want to calculate the discriminant.

        Returns:
            A ndarray with the log likelihoods corresponding to the
            output classes.
        """
        # Calculates difference wrt. the mean (x - un)
        mean_values = np.array([m.data_matrix[0] for m in self.means_])

        X_centered = (
            X[:, np.newaxis, :, :]
            - mean_values[np.newaxis, :, :, :]
        )

        # Calculates mahalanobis distance (-1/2*(x - un).T*inv(sum)*(x - un))
        mahalanobis_distances = np.reshape(
            np.transpose(X_centered, axes=(0, 1, 3, 2))
            @ np.linalg.solve(
                self._regularized_covariances,
                X_centered,
            ),
            (-1, self.classes_.size),
        )

        return np.asarray(
            (
                -0.5 * self._log_determinant_covariances
                - 0.5 * mahalanobis_distances
                + self._log_priors
            ),
        )



"""Representation of functional data."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "basis",
        "extrapolation",
        "grid",
        "interpolation",
    ],
    submod_attrs={
        '_functional_data': ["FData", "concatenate"],
        'basis': ["FDataBasis"],
        'grid': ["FDataGrid"],
    },
)

if TYPE_CHECKING:
    from ._functional_data import FData as FData, concatenate as concatenate
    from .basis import FDataBasis as FDataBasis
    from .grid import FDataGrid as FDataGrid



"""Feature extraction."""
import warnings

from .. import FPCA

warnings.warn(
    'The module "feature_extraction" is deprecated.'
    'Please use "dim_reduction" for FPCA'
    'or "feature_construction" for feature construction techniques',
    category=DeprecationWarning,
    stacklevel=2,
)



"""Base classes for the neighbor estimators."""
from __future__ import annotations

import copy
from typing import Any, Callable, Generic, Tuple, TypeVar, Union, overload

import numpy as np
import sklearn.neighbors
from scipy.sparse import csr_matrix
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted
from typing_extensions import Literal

from skfda.misc.metrics._utils import PairwiseMetric

from .._utils._sklearn_adapter import (
    BaseEstimator,
    ClassifierMixin,
    RegressorMixin,
)
from .._utils._utils import _classifier_get_classes
from ..misc.metrics import l2_distance
from ..misc.metrics._utils import _fit_metric
from ..representation import FData, concatenate
from ..typing._metric import Metric
from ..typing._numpy import NDArrayFloat, NDArrayInt, NDArrayStr

FDataType = TypeVar("FDataType", bound="FData")
SelfType = TypeVar("SelfType", bound="NeighborsBase[Any, Any]")
SelfTypeClassifier = TypeVar(
    "SelfTypeClassifier",
    bound="NeighborsClassifierMixin[Any, Any]",
)
SelfTypeRegressor = TypeVar(
    "SelfTypeRegressor",
    bound="NeighborsRegressorMixin[Any, Any]",
)
Input = TypeVar("Input", contravariant=True, bound=Union[NDArrayFloat, FData])
Target = TypeVar("Target")
TargetClassification = TypeVar(
    "TargetClassification",
    bound=Union[NDArrayInt, NDArrayStr],
)
TargetRegression = TypeVar(
    "TargetRegression",
    bound=Union[NDArrayFloat, FData],
)
TargetRegressionMultivariate = TypeVar(
    "TargetRegressionMultivariate",
    bound=NDArrayFloat,
)
TargetRegressionFData = TypeVar(
    "TargetRegressionFData",
    bound=FData,
)

WeightsType = Union[
    Literal["uniform", "distance"],
    Callable[[NDArrayFloat], NDArrayFloat],
]
AlgorithmType = Literal["auto", "ball_tree", "kd_tree", "brute"]


class NeighborsBase(BaseEstimator, Generic[Input, Target]):
    """Base class for nearest neighbors estimators."""

    def __init__(
        self,
        n_neighbors: int | None = None,
        radius: float | None = None,
        weights: WeightsType = "uniform",
        algorithm: AlgorithmType = "auto",
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ):
        self.n_neighbors = n_neighbors
        self.radius = radius
        self.weights = weights
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.metric = metric
        self.n_jobs = n_jobs

    def _check_is_fitted(self) -> None:
        """
        Check if the estimator is fitted.

        Raises:
            NotFittedError: If the estimator is not fitted.

        """
        sklearn_check_is_fitted(self, ['_estimator'])

    def fit(
        self: SelfType,
        X: Input,
        y: Target,
    ) -> SelfType:
        """
        Fit the model using X as training data and y as target values.

        Args:
            X: Training data. FDataGrid with the training data or array matrix
                with shape [n_samples, n_samples] if metric='precomputed'.
            y: Target values of shape = [n_samples] or [n_samples, n_outputs].
                In the case of unsupervised search, this parameter is ignored.

        Returns:
            Self.

        Note:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        return self._fit(X, y)

    def _fit(
        self: SelfType,
        X: Input,
        y: Target,
        fit_with_zeros: bool = True,
    ) -> SelfType:
        # If metric is precomputed no diferences with the Sklearn estimator
        self._estimator = self._init_estimator()

        self._fitted_with_distances = True

        if self.metric == 'precomputed':
            if isinstance(y, FData):  # For functional response regression
                self._fit_y: Target = copy.deepcopy(y)
            self._estimator.fit(X, y)
        else:
            _fit_metric(self.metric, X)
            self._fit_X = copy.deepcopy(X)
            self._fit_y = copy.deepcopy(y)
            if fit_with_zeros:
                self._fitted_with_distances = False
                self._estimator.fit(np.zeros(shape=(len(X), len(X))), y)
            else:
                distances = PairwiseMetric(self.metric)(X)
                self._estimator.fit(distances, y)

        return self

    def _refit_with_distances(self) -> None:
        if not self._fitted_with_distances:
            assert self.metric != "precomputed"
            distances = PairwiseMetric(self.metric)(self._fit_X)
            self._estimator.fit(distances, self._fit_y)
            self._fitted_with_distances = True

    def _X_to_distances(  # noqa: N802
        self,
        X: Input,
    ) -> NDArrayFloat:

        if self.metric == 'precomputed':
            return X

        return PairwiseMetric(self.metric)(X, self._fit_X)

    def _init_estimator(
        self,
    ) -> sklearn.neighbors.NearestNeighbors:
        """Initialize the sklearn nearest neighbors estimator."""
        return sklearn.neighbors.NearestNeighbors(
            n_neighbors=self.n_neighbors,
            radius=self.radius,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )


class KNeighborsMixin(NeighborsBase[Input, Target]):
    """Mixin class for K-Neighbors."""

    @overload
    def kneighbors(
        self,
        X: Input | None = None,
        n_neighbors: int | None = None,
        *,
        return_distance: Literal[True] = True,
    ) -> Tuple[NDArrayFloat, NDArrayInt]:
        pass

    @overload
    def kneighbors(
        self,
        X: Input | None = None,
        n_neighbors: int | None = None,
        *,
        return_distance: Literal[False],
    ) -> NDArrayInt:
        pass

    def kneighbors(
        self,
        X: Input | None = None,
        n_neighbors: int | None = None,
        *,
        return_distance: bool = True,
    ) -> NDArrayInt | Tuple[NDArrayFloat, NDArrayInt]:
        """
        Find the K-neighbors of a point.

        Returns indices of and distances to the neighbors of each point.

        Args:
            X: FDatagrid with the query functions or  matrix
                (n_query, n_indexed) if metric == 'precomputed'. If not
                provided, neighbors of each indexed point are returned. In
                this case, the query point is not considered its own neighbor.
            n_neighbors: Number of neighbors to get (default is the value
                passed to the constructor).
            return_distance: Defaults to True. If False,
                distances will not be returned.

        Returns:
            dist : array
                Array representing the lengths to points, only present if
                return_distance=True
            ind : array
                Indices of the nearest points in the population matrix.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors()
            >>> neigh.fit(fd)
            NearestNeighbors(...)

            Now we can query the k-nearest neighbors.

            >>> distances, index = neigh.kneighbors(fd[:2])
            >>> index # Index of k-neighbors of samples 0 and 1
            array([[ 0,  7,  6, 11,  2],...)

            >>> distances.round(2) # Distances to k-neighbors
            array([[ 0.  ,  0.28,  0.29,  0.29,  0.3 ],
                   [ 0.  ,  0.27,  0.28,  0.29,  0.3 ]])

        Notes:
            This method wraps the corresponding sklearn routine in the
            module ``sklearn.neighbors``.

        """
        self._check_is_fitted()
        if X is None:
            self._refit_with_distances()

        X_dist = None if X is None else self._X_to_distances(X)

        return self._estimator.kneighbors(  # type: ignore [no-any-return]
            X_dist,
            n_neighbors,
            return_distance,
        )

    def kneighbors_graph(
        self,
        X: Input | None = None,
        n_neighbors: int | None = None,
        mode: Literal["connectivity", "distance"] = "connectivity",
    ) -> csr_matrix:
        """
        Compute the (weighted) graph of k-Neighbors for points in X.

        Args:
            X: FDatagrid with the query functions or  matrix
                (n_query, n_indexed) if metric == 'precomputed'. If not
                provided, neighbors of each indexed point are returned. In
                this case, the query point is not considered its own neighbor.
            n_neighbors: Number of neighbors to get (default is the value
                passed to the constructor).
            mode: Type of returned matrix: 'connectivity' will return the
                connectivity matrix with ones and zeros, in 'distance' the
                edges are distance between points.

        Returns:
            Sparse matrix in CSR format, shape = [n_samples, n_samples_fit]
            n_samples_fit is the number of samples in the fitted data
            A[i, j] is assigned the weight of edge that connects i to j.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator.

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors()
            >>> neigh.fit(fd)
            NearestNeighbors(...)

            Now we can obtain the graph of k-neighbors of a sample.

            >>> graph = neigh.kneighbors_graph(fd[0])
            >>> print(graph)
              (0, 0)	1.0
              (0, 7)	1.0
              (0, 6)	1.0
              (0, 11)	1.0
              (0, 2)	1.0

        Notes:
            This method wraps the corresponding sklearn routine in the
            module ``sklearn.neighbors``.

        """
        self._check_is_fitted()
        if X is None:
            self._refit_with_distances()

        X_dist = None if X is None else self._X_to_distances(X)

        return self._estimator.kneighbors_graph(X_dist, n_neighbors, mode)


class RadiusNeighborsMixin(NeighborsBase[Input, Target]):
    """Mixin Class for Raius Neighbors."""

    @overload
    def radius_neighbors(
        self,
        X: Input | None = None,
        radius: float | None = None,
        *,
        return_distance: Literal[True] = True,
    ) -> Tuple[NDArrayFloat, NDArrayInt]:  # TODO: Fix return type
        pass

    @overload
    def radius_neighbors(
        self,
        X: Input | None = None,
        radius: float | None = None,
        *,
        return_distance: Literal[False],
    ) -> NDArrayInt:
        pass

    def radius_neighbors(
        self,
        X: Input | None = None,
        radius: float | None = None,
        *,
        return_distance: bool = True,
    ) -> NDArrayInt | Tuple[NDArrayFloat, NDArrayInt]:  # TODO: Fix return type
        """
        Find the neighbors within a given radius of a fdatagrid.

        Return the indices and distances of each point from the dataset
        lying in a ball with size ``radius`` around the points of the query
        array. Points lying on the boundary are included in the results.
        The result points are *not* necessarily sorted by distance to their
        query point.

        Args:
            X: Sample or samples whose neighbors will be returned. If not
                provided, neighbors of each indexed point are returned. In this
                case, the query point is not considered its own neighbor.
            radius: Limiting distance of neighbors to return.
                (default is the value passed to the constructor).
            return_distance: Defaults to True. If False, distances will not be
                returned.

        Returns:
            (array, shape (n_samples): dist : array of arrays representing the
                distances to each point, only present if return_distance=True.
                The distance values are computed according to the ``metric``
                constructor parameter.
            (array, shape (n_samples,): An array of arrays of indices of the
                approximate nearest points from the population matrix that lie
                within a ball of size ``radius`` around the query points.

        Examples:
            Firstly, we will create a toy dataset.

            >>> from skfda.datasets import make_sinusoidal_process
            >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
            >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
            ...                               phase_std=.25, random_state=0)
            >>> fd = fd1.concatenate(fd2)

            We will fit a Nearest Neighbors estimator.

            >>> from skfda.ml.clustering import NearestNeighbors
            >>> neigh = NearestNeighbors(radius=.3)
            >>> neigh.fit(fd)
            NearestNeighbors(...radius=0.3...)

            Now we can query the neighbors in the radius.

            >>> distances, index = neigh.radius_neighbors(fd[:2])
            >>> index[0] # Neighbors of sample 0
            array([ 0,  2,  6,  7, 11]...)

            >>> distances[0].round(2) # Distances to neighbors of the sample 0
            array([ 0.  ,  0.3 ,  0.29,  0.28,  0.29])


        See also:
            kneighbors

        Notes:
            Because the number of neighbors of each point is not necessarily
            equal, the results for multiple query points cannot be fit in a
            standard data array.
            For efficiency, `radius_neighbors` returns arrays of objects, where
            each object is a 1D array of indices or distances.

            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        self._check_is_fitted()
        if X is None:
            self._refit_with_distances()

        X_dist = None if X is None else self._X_to_distances(X)

        return (  # type: ignore [no-any-return]
            self._estimator.radius_neighbors(
                X_dist,
                radius=radius,
                return_distance=return_distance,
            )
        )

    def radius_neighbors_graph(
        self,
        X: Input | None = None,
        radius: float | None = None,
        mode: Literal["connectivity", "distance"] = 'connectivity',
    ) -> csr_matrix:
        """
        Compute the (weighted) graph of Neighbors for points in X.

        Neighborhoods are restricted the points at a distance lower than
        radius.

        Args:
            X:  The query sample or samples. If not provided, neighbors of
                each indexed point are returned. In this case, the query
                point is not considered its own neighbor.
            radius: Radius of neighborhoods. (default is the value passed
                to the constructor).
            mode: Type of returned matrix: 'connectivity' will return the
                connectivity matrix with ones and zeros, in 'distance'
                the edges are distance between points.

        Returns:
            sparse matrix in CSR format, shape = [n_samples, n_samples]
            A[i, j] is assigned the weight of edge that connects i to j.

        Notes:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        self._check_is_fitted()
        if X is None:
            self._refit_with_distances()

        X_dist = None if X is None else self._X_to_distances(X)

        return self._estimator.radius_neighbors_graph(
            X_dist,
            radius=radius,
            mode=mode,
        )


class NeighborsClassifierMixin(
    NeighborsBase[Input, TargetClassification],
    ClassifierMixin[Input, TargetClassification],
):
    """Mixin class for classifiers based in nearest neighbors."""

    def predict(
        self,
        X: Input,
    ) -> TargetClassification:
        """
        Predict the class labels for the provided data.

        Args:
            X: Test samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns:
            Array of shape [n_samples] or [n_samples, n_outputs] with class
            labels for each data sample.

        Notes:
            This method wraps the corresponding sklearn routine in the module
            ``sklearn.neighbors``.

        """
        self._check_is_fitted()

        X_dist = self._X_to_distances(X)

        return self._estimator.predict(X_dist)  # type: ignore [no-any-return]

    def predict_proba(
        self,
        X: Input,
    ) -> NDArrayFloat:
        """
        Calculate probability estimates for the test data X.

        Args:
            X: FDataGrid with the test samples or array (n_query, n_indexed)
                if metric == 'precomputed'.

        Returns:
            The class probabilities of the input samples. Classes are
            ordered by lexicographic order.

        """
        self._check_is_fitted()

        X_dist = self._X_to_distances(X)

        return (  # type: ignore [no-any-return]
            self._estimator.predict_proba(X_dist)
        )

    def fit(
        self: SelfTypeClassifier,
        X: Input,
        y: TargetClassification,
    ) -> SelfTypeClassifier:
        """
        Fit the model using X as training data and y as responses.

        Args:
            X: Training data. FDataGrid
                with the training data or array matrix with shape
                [n_samples, n_samples] if metric='precomputed'.
            y: Training data. FData with the training respones (functional
                response case) or array matrix with length `n_samples` in
                the multivariate response case.

        Returns:
            Self.

        Note:
            This method adds the attribute `classes_` to the classifier.
        """
        self.classes_ = _classifier_get_classes(y)[0]
        return super().fit(X, y)


class NeighborsRegressorMixin(
    NeighborsBase[Input, TargetRegression],
    RegressorMixin[Input, TargetRegression],
):
    """Mixin class for the regressors based on neighbors."""

    def _average(
        self,
        X: TargetRegression,
        weights: NDArrayFloat | None = None,
    ) -> TargetRegression:
        """Compute weighted average."""
        if weights is None:
            return np.mean(X, axis=0)  # type: ignore [no-any-return]

        weights /= np.sum(weights)

        return np.sum(X * weights, axis=0)  # type: ignore [no-any-return]

    def _prediction_from_neighbors(
        self,
        neighbors: TargetRegression,
        distance: NDArrayFloat,
    ) -> TargetRegression:

        if self.weights == 'uniform':
            weights = None
        elif self.weights == 'distance':
            weights = self._distance_weights(distance)
        else:
            weights = self.weights(distance)

        return self._average(neighbors, weights)

    def fit(
        self: SelfTypeRegressor,
        X: Input,
        y: TargetRegression,
    ) -> SelfTypeRegressor:
        """
        Fit the model using X as training data and y as responses.

        Args:
            X: Training data. FDataGrid
                with the training data or array matrix with shape
                [n_samples, n_samples] if metric='precomputed'.
            y: Training data. FData with the training respones (functional
                response case) or array matrix with length `n_samples` in
                the multivariate response case.

        Returns:
            Self.

        """
        self._functional = isinstance(y, FData)
        return super().fit(X, y)

    def _distance_weights(
        self,
        distance: NDArrayFloat,
    ) -> NDArrayFloat:
        """Return weights based on distance reciprocal."""
        idx = (distance == 0)
        if np.any(idx):
            weights = distance
            weights[idx] = 1
            weights[~idx] = 0
        else:
            weights = 1 / distance

        return weights

    def predict(
        self,
        X: Input,
    ) -> TargetRegression:
        """
        Predict the target for the provided data.

        Args:
            X: FDataGrid with the test
                samples or array (n_query, n_indexed) if metric ==
                'precomputed'.

        Returns:
            array of shape = [n_samples] or [n_samples, n_outputs]
            or :class:`FData` containing as many samples as X.

        """
        self._check_is_fitted()

        # Choose type of prediction
        if self._functional:
            return self._functional_predict(X)

        return self._multivariate_predict(X)

    def _multivariate_predict(
        self: NeighborsRegressorMixin[Input, TargetRegressionMultivariate],
        X: Input,
    ) -> TargetRegressionMultivariate:
        """Predict a multivariate target."""
        X_dist = self._X_to_distances(X)

        return self._estimator.predict(X_dist)  # type: ignore [no-any-return]

    def _functional_predict(
        self: NeighborsRegressorMixin[FData, Any],
        X: Input,
    ) -> TargetRegression:
        """Predict functional responses."""
        distances, neighbors = self._query(X)

        iterable = (
            self._fit_y.dtype._na_repr()  # noqa: WPS437
            if len(idx) == 0
            else self._prediction_from_neighbors(
                self._fit_y[idx],
                dist,
            )
            for idx, dist in zip(neighbors, distances)
        )

        return concatenate(iterable)  # type: ignore[no-any-return]



from __future__ import annotations

from typing import TypeVar

from sklearn.utils.validation import check_is_fitted

from ..._utils._sklearn_adapter import BaseEstimator, RegressorMixin
from ...misc.regularization import L2Regularization
from ...preprocessing.dim_reduction import FPCA
from ...representation import FData
from ...representation.basis import Basis, CustomBasis, FDataBasis
from ...typing._numpy import NDArrayFloat
from ._linear_regression import LinearRegression

FPCARegressionSelf = TypeVar("FPCARegressionSelf", bound="FPCARegression")


class FPCARegression(
    BaseEstimator,
    RegressorMixin,
):
    r"""Regression using Functional Principal Components Analysis.

    It performs Functional Principal Components Analysis to reduce the
    dimension of the functional data, and then uses a linear regression model
    to relate the transformed data to a scalar value.

    Args:
        n_components: Number of principal components to keep. Defaults to 5.
        fit\_intercept: If True, the linear model is calculated with an
            intercept.  Defaults to ``True``.
        pca_regularization: Regularization parameter for the principal
            component extraction. If None then no regularization is applied.
            Defaults to ``None``.
        regression_regularization: Regularization parameter for the linear
            regression. If None then no regularization is applied.
            Defaults to ``None``.
        components_basis: Basis used for the principal components. If None
            then the basis of the input data is used. Defaults to None.
            It is only used if the input data is a FDataBasis object.

    Attributes:
        n\_components\_: Number of principal components used.
        components\_: Principal components.
        coef\_: Coefficients of the linear regression model.
        explained\_variance\_: Amount of variance explained by
            each of the selected components.
        explained\_variance\_ratio\_: Percentage of variance
            explained by each of the selected components.

    Examples:
        Using the Berkeley Growth Study dataset, we can fit the model.

        >>> import skfda
        >>> dataset = skfda.datasets.fetch_growth()
        >>> fd = dataset["data"]
        >>> y = dataset["target"]
        >>> reg = skfda.ml.regression.FPCARegression(n_components=2)
        >>> reg.fit(fd, y)
        FPCARegression(n_components=2)

        Then, we can predict the target values and calculate the
        score.

        >>> score = reg.score(fd, y)
        >>> reg.predict(fd) # doctest:+ELLIPSIS
        array([...])

    """

    def __init__(
        self,
        n_components: int = 5,
        fit_intercept: bool = True,
        pca_regularization: L2Regularization | None = None,
        regression_regularization: L2Regularization | None = None,
        components_basis: Basis | None = None,
    ) -> None:
        self.n_components = n_components
        self.fit_intercept = fit_intercept
        self.pca_regularization = pca_regularization
        self.regression_regularization = regression_regularization
        self.components_basis = components_basis

    def fit(
        self,
        X: FData,
        y: NDArrayFloat,
    ) -> FPCARegressionSelf:
        """Fit the model according to the given training data.

        Args:
            X: Functional data.
            y: Target values.

        Returns:
            self

        """
        self._fpca = FPCA(
            n_components=self.n_components,
            centering=True,
            regularization=self.pca_regularization,
            components_basis=self.components_basis,
        )
        self._linear_model = LinearRegression(
            fit_intercept=self.fit_intercept,
            regularization=self.regression_regularization,
        )
        transformed_coefficients = self._fpca.fit_transform(X)

        # The linear model is fitted with the observations expressed in the
        # basis of the principal components.
        self.fpca_basis = CustomBasis(
            fdata=self._fpca.components_,
        )

        X_transformed = FDataBasis(
            basis=self.fpca_basis,
            coefficients=transformed_coefficients,
        )
        self._linear_model.fit(X_transformed, y)

        self.n_components_ = self.n_components
        self.components_ = self._fpca.components_
        self.coef_ = self._linear_model.coef_
        self.explained_variance_ = self._fpca.explained_variance_
        self.explained_variance_ratio_ = self._fpca.explained_variance_ratio_

        return self

    def predict(
        self,
        X: FData,
    ) -> NDArrayFloat:
        """Predict using the linear model.

        Args:
            X: Functional data.

        Returns:
            Target values.

        """
        check_is_fitted(self)

        X_transformed = FDataBasis(
            basis=self.fpca_basis,
            coefficients=self._fpca.transform(X),
        )

        return self._linear_model.predict(X_transformed)



"""Smoothing."""
import warnings
from typing import TYPE_CHECKING, Any

import lazy_loader as lazy

_normal_getattr, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "validation",
    ],
    submod_attrs={
        "_basis": ["BasisSmoother"],
        "_kernel_smoothers": ["KernelSmoother"],
    },
)

if TYPE_CHECKING:
    from ._basis import BasisSmoother as BasisSmoother
    from ._kernel_smoothers import KernelSmoother as KernelSmoother

__kernel_smoothers__imported__ = False


def __getattr__(name: str) -> Any:
    global __kernel_smoothers__imported__
    if name == "kernel_smoothers" and not __kernel_smoothers__imported__:
        __kernel_smoothers__imported__ = True
        from . import kernel_smoothers
        return kernel_smoothers

    return _normal_getattr(name)



"""Dim reduction."""
from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Any

import lazy_loader as lazy

_normal_getattr, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "variable_selection",
    ],
    submod_attrs={
        "_fpca": ["FPCA"],
        "_neighbor_transforms": ["KNeighborsTransformer"]
    },
)

if TYPE_CHECKING:
    from ._fpca import FPCA as FPCA
    from ._neighbor_transforms import KNeighborsTransformer as KNeighborsTransformer


def __getattr__(name: str) -> Any:
    if name in {"projection", "feature_extraction"}:
        return importlib.import_module(f".{name}", __name__)
    return _normal_getattr(name)



"""Outlier detection methods."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        '_boxplot': ["BoxplotOutlierDetector"],
        "_directional_outlyingness": [
            "MSPlotOutlierDetector",
            "directional_outlyingness_stats",
        ],
        "_outliergram": ["OutliergramOutlierDetector"],
        "neighbors_outlier": ["LocalOutlierFactor"],
    },
)

if TYPE_CHECKING:
    from ._boxplot import BoxplotOutlierDetector as BoxplotOutlierDetector
    from ._directional_outlyingness import (
        MSPlotOutlierDetector as MSPlotOutlierDetector,
        directional_outlyingness_stats as directional_outlyingness_stats,
    )
    from ._outliergram import (
        OutliergramOutlierDetector as OutliergramOutlierDetector
    )
    from .neighbors_outlier import LocalOutlierFactor as LocalOutlierFactor



"""Functional data descriptive statistics."""
from __future__ import annotations

from builtins import isinstance
from typing import Callable, TypeVar, Union

import numpy as np
from scipy import integrate
from scipy.stats import rankdata

from skfda._utils.ndfunction import average_function_value

from ...misc.metrics._lp_distances import l2_distance
from ...representation import FData, FDataGrid
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat
from ..depth import Depth, ModifiedBandDepth

F = TypeVar('F', bound=FData)
T = TypeVar('T', bound=Union[NDArrayFloat, FData])


def mean(
    X: F,
    weights: NDArrayFloat | None = None,
) -> F:
    """
    Compute the mean of all the samples in a FData object.

    Args:
        X: Object containing all the samples whose mean is wanted.
        weights: Sample weight. By default, uniform weight are used.

    Returns:
        Mean of all the samples in the original object, as a
        :term:`functional data object` with just one sample.

    """
    if weights is None:
        return X.mean()

    weight = (1 / np.sum(weights)) * weights
    return (X * weight).sum()


def var(X: FData, correction: int = 0) -> FDataGrid:
    """
    Compute the variance of a set of samples in a FData object.

    Args:
        X: Object containing all the set of samples whose variance is desired.
        correction: degrees of freedom adjustment. The divisor used in the
            calculation is `N - correction`, where `N` represents the number of
            elements. Default: `0`.

    Returns:
        Variance of all the samples in the original object, as a
        :term:`functional data object` with just one sample.

    """
    return X.var(correction=correction)  # type: ignore[no-any-return]


def gmean(X: FDataGrid) -> FDataGrid:
    """
    Compute the geometric mean of all the samples in a FDataGrid object.

    Args:
        X: Object containing all the samples whose geometric mean is wanted.

    Returns:
        Geometric mean of all the samples in the original object, as a
        :term:`functional data object` with just one sample.

    """
    return X.gmean()


def cov(
    X: FData,
    correction: int = 0,
) -> Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]:
    """
    Compute the covariance.

    Calculates the covariance matrix representing the covariance of the
    functional samples at the observation points.

    Args:
        X: Object containing different samples of a functional variable.
        correction: degrees of freedom adjustment. The divisor used in the
            calculation is `N - correction`, where `N` represents the number of
            elements. Default: `0`.


    Returns:
        Covariance of all the samples in the original object, as a
        callable.

    """
    return X.cov(correction=correction)


def modified_epigraph_index(X: FDataGrid) -> NDArrayFloat:
    """
    Calculate the Modified Epigraph Index of a FDataGrid.

    The MEI represents the mean time a curve stays below other curve.
    In this case we will calculate the MEI for each curve in relation
    with all the other curves of our dataset.

    """
    # Functions containing at each point the number of curves
    # are above it.
    num_functions_above = X.copy(
        data_matrix=rankdata(
            -X.data_matrix,
            method='max',
            axis=0,
        ) - 1,
    )

    return (
        average_function_value(num_functions_above)
        / num_functions_above.n_samples
    ).ravel()


def depth_based_median(
    X: T,
    depth_method: Depth[T] | None = None,
) -> T:
    """
    Compute the median based on a depth measure.

    The depth based median is the deepest curve given a certain
    depth measure.

    Args:
        X: Object containing different samples of a
            functional variable.
        depth_method: Depth method used to order the data. Defaults to
            :func:`modified band
            depth <skfda.exploratory.depth.ModifiedBandDepth>`.

    Returns:
        Object containing the computed depth_based median.

    See also:
        :func:`geometric_median`

    """
    depth_method_used: Depth[T]

    if depth_method is None:
        assert isinstance(X, FDataGrid)
        depth_method_used = ModifiedBandDepth()
    else:
        depth_method_used = depth_method

    depth = depth_method_used(X)
    indices_descending_depth = (-depth).argsort(axis=0)

    # The median is the deepest curve
    return X[indices_descending_depth[0]]


def _weighted_average(X: T, weights: NDArrayFloat) -> T:

    if isinstance(X, FData):
        return (X * weights).sum()

    return (X.T * weights).T.sum(axis=0)  # type: ignore[no-any-return]


def geometric_median(
    X: T,
    *,
    tol: float = 1.e-8,
    metric: Metric[T] = l2_distance,
) -> T:
    r"""
    Compute the geometric median.

    The sample geometric median is the point that minimizes the :math:`L_1`
    norm of the vector of distances to all observations:

    .. math::

        \underset{y \in L(\mathcal{T})}{\arg \min}
        \sum_{i=1}^N \left \| x_i-y \right \|

    The geometric median in the functional case is also described in
    :footcite:`gervini_2008_robust`.
    Instead of the proposed algorithm, however, the current implementation
    uses the corrected Weiszfeld algorithm to compute the median.

    Args:
        X: Object containing different samples of a
            functional variable.
        tol: tolerance used to check convergence.
        metric: metric used to compute the vector of distances. By
            default is the :math:`L_2` distance.

    Returns:
        Object containing the computed geometric median.

    Example:

        >>> from skfda import FDataGrid
        >>> data_matrix = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
        >>> X = FDataGrid(data_matrix)
        >>> median = geometric_median(X)
        >>> median.data_matrix[0, ..., 0]
        array([ 1. ,  1. ,  3. ,  0.5])

    See also:
        :func:`depth_based_median`

    References:
        .. footbibliography::

    """
    weights = np.full(len(X), 1 / len(X))
    median = _weighted_average(X, weights)
    distances = metric(X, median)

    while True:
        zero_distances = (distances == 0)
        n_zeros = np.sum(zero_distances)
        weights_new = (
            (1 / distances) / np.sum(1 / distances) if n_zeros == 0
            else (1 / n_zeros) * zero_distances
        )

        median_new = _weighted_average(X, weights_new)

        if l2_distance(median_new, median) < tol:
            return median_new

        distances = metric(X, median_new)

        weights, median = (weights_new, median_new)


def trim_mean(
    X: F,
    proportiontocut: float,
    *,
    depth_method: Depth[F] | None = None,
) -> FDataGrid:
    """Compute the trimmed means based on a depth measure.

    The trimmed means consists in computing the mean function without a
    percentage of least deep curves. That is, we first remove the least deep
    curves and then we compute the mean as usual.

    Note that in scipy the leftmost and rightmost proportiontocut data are
    removed. In this case, as we order the data by the depth, we only remove
    those that have the least depth values.

    Args:
        X: Object containing different samples of a
            functional variable.
        proportiontocut: Indicates the percentage of functions to
            remove. It is not easy to determine as it varies from dataset to
            dataset.
        depth_method: Method used to order the data. Defaults to
            :func:`modified band depth
            <skfda.exploratory.depth.ModifiedBandDepth>`.

    Returns:
        Object containing the computed trimmed mean.

    """
    if depth_method is None:
        depth_method = ModifiedBandDepth()

    n_samples_to_keep = (len(X) - int(len(X) * proportiontocut))

    # compute the depth of each curve and store the indexes in descending order
    depth = depth_method(X)
    indices_descending_depth = (-depth).argsort(axis=0)

    trimmed_curves = X[indices_descending_depth[:n_samples_to_keep]]

    return trimmed_curves.mean()



from __future__ import annotations

from typing import Callable, TypeVar

from ....representation import FData
from ....typing._numpy import NDArrayFloat
from ._base import CovarianceEstimator

Input = TypeVar("Input", bound=FData)


class EmpiricalCovariance(
    CovarianceEstimator[Input],
):
    covariance_: Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]

    def fit(self, X: Input, y: object = None) -> EmpiricalCovariance[Input]:
        super().fit(X, y)
        self.covariance_ = X.cov()
        return self



"""Test smoothing methods."""
import unittest
from typing import Tuple

import numpy as np
import sklearn
from sklearn.datasets import load_digits
from typing_extensions import Literal

import skfda
import skfda.preprocessing.smoothing as smoothing
import skfda.preprocessing.smoothing.validation as validation
from skfda._utils import _check_estimator
from skfda.datasets import fetch_weather
from skfda.misc.hat_matrix import (
    HatMatrix,
    KNeighborsHatMatrix,
    LocalLinearRegressionHatMatrix,
    NadarayaWatsonHatMatrix,
)
from skfda.misc.operators import LinearDifferentialOperator
from skfda.misc.regularization import L2Regularization
from skfda.preprocessing.smoothing import KernelSmoother
from skfda.representation.basis import BSplineBasis, MonomialBasis
from skfda.representation.grid import FDataGrid


class TestSklearnEstimators(unittest.TestCase):
    """Test for sklearn estimators."""

    def test_kernel_smoothing(self) -> None:
        """Test if estimator adheres to scikit-learn conventions."""
        _check_estimator(KernelSmoother)


class _LinearSmootherLeaveOneOutScorerAlternative:
    """Alternative implementation of the LinearSmootherLeaveOneOutScorer."""

    def __call__(
        self,
        estimator: KernelSmoother,
        X: FDataGrid,
        y: FDataGrid,
    ) -> float:
        """Calculate Leave-One-Out score."""
        estimator_clone = sklearn.base.clone(estimator)

        estimator_clone._cv = True  # noqa: WPS437
        y_est = estimator_clone.fit_transform(X)

        return float(
            -np.mean((y.data_matrix[..., 0] - y_est.data_matrix[..., 0])**2),
        )


class TestLeaveOneOut(unittest.TestCase):
    """Tests of Leave-One-Out score for kernel smoothing."""

    def _test_generic(
        self,
        estimator: KernelSmoother,
        smoothing_param_name: str = 'kernel_estimator__bandwidth',
    ) -> None:
        loo_scorer = validation.LinearSmootherLeaveOneOutScorer()
        loo_scorer_alt = _LinearSmootherLeaveOneOutScorerAlternative()

        x = np.linspace(-2, 2, 5)
        fd = skfda.FDataGrid(x ** 2, x)

        grid = validation.SmoothingParameterSearch(
            estimator,
            [2, 3],
            param_name=smoothing_param_name,
            scoring=loo_scorer,
        )

        grid.fit(fd)
        score = np.array(grid.cv_results_['mean_test_score'])

        grid_alt = validation.SmoothingParameterSearch(
            estimator,
            [2, 3],
            param_name=smoothing_param_name,
            scoring=loo_scorer_alt,
        )

        grid_alt.fit(fd)
        score_alt = np.array(grid_alt.cv_results_['mean_test_score'])

        np.testing.assert_array_almost_equal(score, score_alt)

    def test_nadaraya_watson(self) -> None:
        """Test Leave-One-Out with Nadaraya Watson method."""
        self._test_generic(KernelSmoother(
            kernel_estimator=NadarayaWatsonHatMatrix(),
        ),
        )

    def test_local_linear_regression(self) -> None:
        """Test Leave-One-Out with Local Linear Regression method."""
        self._test_generic(KernelSmoother(
            kernel_estimator=LocalLinearRegressionHatMatrix(),
        ),
        )

    def test_knn(self) -> None:
        """Test Leave-One-Out with KNNeighbours method."""
        self._test_generic(
            KernelSmoother(
                kernel_estimator=KNeighborsHatMatrix(),
            ),
            smoothing_param_name='kernel_estimator__n_neighbors',
        )


class TestKernelSmoother(unittest.TestCase):
    """Test Kernel Smoother.

    Comparison of results with fda.usc R library
    """

    def _test_hat_matrix(
        self,
        kernel_estimator: HatMatrix,
    ) -> np.typing.NDArray[np.float_]:
        return KernelSmoother(  # noqa: WPS437
            kernel_estimator=kernel_estimator,
        )._hat_matrix(
            input_points=[[1, 2, 3, 4, 5]],
            output_points=[[1, 2, 3, 4, 5]],
        )

    def test_nw(self) -> None:
        """Comparison of NW hat matrix with the one obtained from fda.usc."""
        hat_matrix = self._test_hat_matrix(
            NadarayaWatsonHatMatrix(bandwidth=10),
        )
        hat_matrix_r = [
            [0.206001865, 0.204974427, 0.201922755, 0.196937264, 0.190163689],
            [0.201982911, 0.202995354, 0.201982911, 0.198975777, 0.194063047],
            [0.198003042, 0.200995474, 0.202002968, 0.200995474, 0.198003042],
            [0.194063047, 0.198975777, 0.201982911, 0.202995354, 0.201982911],
            [0.190163689, 0.196937264, 0.201922755, 0.204974427, 0.206001865],
        ]
        np.testing.assert_allclose(hat_matrix, hat_matrix_r)

    def test_llr(self) -> None:
        """Test LLR."""
        hat_matrix = self._test_hat_matrix(
            LocalLinearRegressionHatMatrix(bandwidth=10),
        )

        # For a straight line the estimated results should coincide with
        # the real values
        # r(x) = 3x + 2
        np.testing.assert_allclose(
            np.dot(hat_matrix, [5, 8, 11, 14, 17]),
            [5, 8, 11, 14, 17],
        )

    def test_knn(self) -> None:
        """Comparison of KNN hat matrix with the one obtained from fda.usc."""
        hat_matrix = self._test_hat_matrix(
            KNeighborsHatMatrix(n_neighbors=2),
        )

        hat_matrix_r = [
            [0.500000000, 0.500000000, 0.000000000, 0.000000000, 0.000000000],
            [0.333333333, 0.333333333, 0.333333333, 0.000000000, 0.000000000],
            [0.000000000, 0.333333333, 0.333333333, 0.333333333, 0.000000000],
            [0.000000000, 0.000000000, 0.333333333, 0.333333333, 0.333333333],
            [0.000000000, 0.000000000, 0.000000000, 0.500000000, 0.500000000],
        ]
        np.testing.assert_allclose(hat_matrix, hat_matrix_r)


class TestSurfaceSmoother(unittest.TestCase):
    """Test that smoothing works on surfaces."""

    def test_knn(self) -> None:
        """Check 2D smoothing using digits dataset."""
        X, y = load_digits(return_X_y=True)
        X = X.reshape(-1, 8, 8)

        fd = skfda.FDataGrid(X)

        ks = KernelSmoother(
            kernel_estimator=KNeighborsHatMatrix(n_neighbors=1))
        fd_trans = ks.fit_transform(fd)

        np.testing.assert_allclose(fd_trans.data_matrix, fd.data_matrix)

        ks = KernelSmoother(
            kernel_estimator=KNeighborsHatMatrix(n_neighbors=3))
        fd_trans = ks.fit_transform(fd)

        res = [
            [0, 1.25, 7.75, 10.5, 8.25, 6.25, 1.5, 0],
            [0, 3.2, 9.6, 10.6, 9.8, 8.4, 5.6, 1.25],
            [0.75, 4.4, 9, 6.4, 4.6, 8.4, 6.4, 2],
            [1, 4.8, 7.8, 2.8, 1.6, 7.2, 6.4, 2],
            [1.25, 4.2, 7.2, 1.6, 2, 7.4, 6.4, 2],
            [1, 4.4, 7.4, 3.4, 4.6, 8.2, 5.4, 1.75],
            [0.5, 4, 7.6, 8.4, 7.6, 6.8, 3.8, 0],
            [0, 2, 8.25, 8.5, 8.25, 5.5, 0, 0],
        ]
        np.testing.assert_allclose(fd_trans.data_matrix[0, ..., 0], res)


class TestBasisSmoother(unittest.TestCase):
    """Test Basis Smoother."""

    def test_cholesky(self) -> None:
        """Test Basis Smoother using BSpline basis and Cholesky method."""
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSplineBasis((0, 1), n_basis=5)

        fd = FDataGrid(data_matrix=x, grid_points=t)

        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=10,
            regularization=L2Regularization(
                LinearDifferentialOperator(2),
            ),
            method='cholesky',
            return_basis=True,
        )

        fd_basis = smoother.fit_transform(fd)

        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.6, 0.47, 0.2, -0.07, -0.2]]),
        )

    def test_qr(self) -> None:
        """Test Basis Smoother using BSpline basis and QR method."""
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = BSplineBasis((0, 1), n_basis=5)

        fd = FDataGrid(data_matrix=x, grid_points=t)

        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=10,
            regularization=L2Regularization(
                LinearDifferentialOperator(2),
            ),
            method='qr',
            return_basis=True,
        )

        fd_basis = smoother.fit_transform(fd)

        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.6, 0.47, 0.2, -0.07, -0.2]]),
        )

    def test_monomial_smoothing(self) -> None:
        """Test Basis Smoother using Monomial basis."""
        # It does not have much sense to apply smoothing in this basic case
        # where the fit is very good but its just for testing purposes
        t = np.linspace(0, 1, 5)
        x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t)
        basis = MonomialBasis(n_basis=4)

        fd = FDataGrid(data_matrix=x, grid_points=t)

        smoother = smoothing.BasisSmoother(
            basis=basis,
            smoothing_parameter=1,
            regularization=L2Regularization(
                LinearDifferentialOperator(2),
            ),
            return_basis=True,
        )

        fd_basis = smoother.fit_transform(fd)

        # These results where extracted from the R package fda
        np.testing.assert_array_almost_equal(
            fd_basis.coefficients.round(2),
            np.array([[0.61, -0.88, 0.06, 0.02]]),
        )

    def test_vector_valued_smoothing(self) -> None:
        """Test Basis Smoother for vector values functions."""
        X, _ = fetch_weather(return_X_y=True)

        basis_dim = skfda.representation.basis.FourierBasis(
            n_basis=7, domain_range=X.domain_range,
        )

        basis = skfda.representation.basis.VectorValuedBasis(
            [basis_dim] * 2,
        )

        method_set: Tuple[Literal['cholesky', 'qr', 'svd'], ...] = (
            'cholesky',
            'qr',
            'svd',
        )
        for method in method_set:
            with self.subTest(method=method):

                basis_smoother = smoothing.BasisSmoother(
                    basis,
                    regularization=L2Regularization(
                        LinearDifferentialOperator(2),
                    ),
                    return_basis=True,
                    smoothing_parameter=1,
                    method=method,
                )

                basis_smoother_dim = smoothing.BasisSmoother(
                    basis_dim,
                    regularization=L2Regularization(
                        LinearDifferentialOperator(2),
                    ),
                    return_basis=True,
                    smoothing_parameter=1,
                    method=method,
                )

                X_basis = basis_smoother.fit_transform(X)

                self.assertEqual(X_basis.dim_codomain, 2)

                self.assertEqual(X_basis.coordinates[0].basis, basis_dim)
                np.testing.assert_allclose(
                    X_basis.coordinates[0].coefficients,
                    basis_smoother_dim.fit_transform(
                        X.coordinates[0],
                    ).coefficients,
                )

                self.assertEqual(X_basis.coordinates[1].basis, basis_dim)
                np.testing.assert_allclose(
                    X_basis.coordinates[1].coefficients,
                    basis_smoother_dim.fit_transform(
                        X.coordinates[1],
                    ).coefficients,
                )



from __future__ import annotations

from typing import Any, TypeVar

from ..._utils import _to_grid_points
from ...typing._base import GridPointsLike
from ...typing._numpy import NDArrayFloat
from ._basis import Basis

T = TypeVar("T", bound="_GridBasis")


class _GridBasis(Basis):
    """
    Basis associated to a grid of points.

    Used to express functions whose values are known in a grid of points.
    Each basis function is guaranteed to be zero in all points of the grid
    except one, where it is one. The intermediate values are interpolated
    depending on the interpolation object passed as argument.

    This basis is used internally as an alternate representation of a
    FDataGrid object. In certain cases, it is more convenient to work with
    FDataGrid objects as a basis representation in this basis since, then,
    all functional data can be represented as an FDataBasis object.

    Parameters:
        grid_points: Grid of points where the functions are evaluated.

    """

    def __init__(
        self,
        *,
        grid_points: GridPointsLike,
    ) -> None:
        """Basis constructor."""
        self.grid_points = _to_grid_points(grid_points)
        domain_range = tuple((s[0], s[-1]) for s in self.grid_points)
        super().__init__(
            domain_range=domain_range,
            n_basis=len(self.grid_points[0]),
        )

    def _evaluate(self, eval_points: NDArrayFloat) -> NDArrayFloat:
        raise NotImplementedError(
            "Evaluation is not implemented in this basis",
        )

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.grid_points == other.grid_points
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}(grid_points={self.grid_points}) "

    def __hash__(self) -> int:
        return hash(
            (
                super(),
                (
                    tuple(grid_point_axis)
                    for grid_point_axis in self.grid_points
                ),
            ),
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload

import sklearn.base

if TYPE_CHECKING:
    from ..typing._numpy import NDArrayFloat, NDArrayInt

SelfType = TypeVar("SelfType")
TransformerNoTarget = TypeVar(
    "TransformerNoTarget",
    bound="TransformerMixin[Any, Any, None]",
)
Input = TypeVar("Input", contravariant=True)
Output = TypeVar("Output", covariant=True)
Target = TypeVar("Target", contravariant=True)
TargetPrediction = TypeVar("TargetPrediction")


class BaseEstimator(  # noqa: D101
    ABC,
    sklearn.base.BaseEstimator,  # type: ignore[misc]
):
    pass  # noqa: WPS604


class TransformerMixin(  # noqa: D101
    ABC,
    Generic[Input, Output, Target],
    sklearn.base.TransformerMixin,  # type: ignore[misc]
):

    @overload
    def fit(
        self: TransformerNoTarget,
        X: Input,
    ) -> TransformerNoTarget:
        pass

    @overload
    def fit(
        self: SelfType,
        X: Input,
        y: Target,
    ) -> SelfType:
        pass

    def fit(  # noqa: D102
        self: SelfType,
        X: Input,
        y: Target | None = None,
    ) -> SelfType:
        return self

    @overload
    def fit_transform(
        self: TransformerNoTarget,
        X: Input,
    ) -> Output:
        pass

    @overload
    def fit_transform(
        self,
        X: Input,
        y: Target,
    ) -> Output:
        pass

    def fit_transform(  # noqa: D102
        self,
        X: Input,
        y: Target | None = None,
        **fit_params: Any,
    ) -> Output:
        if y is None:
            return self.fit(  # type: ignore[no-any-return]
                X,
                **fit_params,
            ).transform(X)

        return self.fit(  # type: ignore[no-any-return]
            X,
            y,
            **fit_params,
        ).transform(X)


class InductiveTransformerMixin(  # noqa: D101
    TransformerMixin[Input, Output, Target],
):

    @abstractmethod
    def transform(  # noqa: D102
        self: SelfType,
        X: Input,
    ) -> Output:
        pass


class OutlierMixin(  # noqa: D101
    ABC,
    Generic[Input],
    sklearn.base.OutlierMixin,  # type: ignore[misc]
):

    def fit_predict(  # noqa: D102
        self,
        X: Input,
        y: object = None,
    ) -> NDArrayInt:
        return self.fit(X, y).predict(X)  # type: ignore[no-any-return]


class ClassifierMixin(  # noqa: D101
    ABC,
    Generic[Input, TargetPrediction],
    sklearn.base.ClassifierMixin,  # type: ignore[misc]
):
    def fit(  # noqa: D102
        self: SelfType,
        X: Input,
        y: TargetPrediction,
    ) -> SelfType:
        return self

    @abstractmethod
    def predict(  # noqa: D102
        self: SelfType,
        X: Input,
    ) -> TargetPrediction:
        pass

    def score(  # noqa: D102
        self,
        X: Input,
        y: Target,
        sample_weight: NDArrayFloat | None = None,
    ) -> float:
        return super().score(  # type: ignore[no-any-return]
            X,
            y,
            sample_weight=sample_weight,
        )


class ClusterMixin(  # noqa: D101
    ABC,
    Generic[Input],
    sklearn.base.ClusterMixin,  # type: ignore[misc]
):
    def fit_predict(  # noqa: D102
        self,
        X: Input,
        y: object = None,
    ) -> NDArrayInt:
        return super().fit_predict(X, y)  # type: ignore[no-any-return]


class RegressorMixin(  # noqa: D101
    ABC,
    Generic[Input, TargetPrediction],
    sklearn.base.RegressorMixin,  # type: ignore[misc]
):
    def fit(  # noqa: D102
        self: SelfType,
        X: Input,
        y: TargetPrediction,
    ) -> SelfType:
        return self

    @abstractmethod
    def predict(  # noqa: D102
        self: SelfType,
        X: Input,
    ) -> TargetPrediction:
        pass

    def score(  # noqa: D102
        self,
        X: Input,
        y: TargetPrediction,
        sample_weight: NDArrayFloat | None = None,
    ) -> float:
        from ..misc.scoring import r2_score
        y_pred = self.predict(X)
        return r2_score(
            y,
            y_pred,
            sample_weight=sample_weight,
        )



"""Initialization module of visualization folder."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "clustering",
        "representation",
    ],
    submod_attrs={
        "_baseplot": ["BasePlot"],
        "_boxplot": ["Boxplot", "SurfaceBoxplot"],
        "_ddplot": ["DDPlot"],
        "_magnitude_shape_plot": ["MagnitudeShapePlot"],
        "_multiple_display": ["MultipleDisplay"],
        "_outliergram": ["Outliergram"],
        "_parametric_plot": ["ParametricPlot"],
        "fpca": ["FPCAPlot"],
    },
)

if TYPE_CHECKING:
    from ._baseplot import BasePlot as BasePlot
    from ._boxplot import Boxplot as Boxplot, SurfaceBoxplot as SurfaceBoxplot
    from ._ddplot import DDPlot as DDPlot
    from ._magnitude_shape_plot import MagnitudeShapePlot as MagnitudeShapePlot
    from ._multiple_display import MultipleDisplay as MultipleDisplay
    from ._outliergram import Outliergram as Outliergram
    from ._parametric_plot import ParametricPlot as ParametricPlot
    from .fpca import FPCAPlot as FPCAPlot



"""Functional Mahalanobis Distance Module."""

from __future__ import annotations

from typing import Callable, Optional, Union

import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from ..._utils._sklearn_adapter import BaseEstimator
from ...representation import FData
from ...representation.basis import Basis
from ...typing._numpy import ArrayLike, NDArrayFloat
from .._math import inner_product_matrix
from ..regularization._regularization import TikhonovRegularization

WeightsCallable = Callable[[np.ndarray], np.ndarray]


class MahalanobisDistance(BaseEstimator):
    """Functional Mahalanobis distance.

    Class that implements functional Mahalanobis distance for both
    basis and grid representations of the data
    :footcite:`berrendero++_2020_mahalanobis`.

    Parameters:
        n_components: Number of eigenvectors to keep from
            functional principal component analysis. Defaults to 10.
        centering: If ``True`` then calculate the mean of the functional
            data object and center the data first. Defaults to ``True``.
        regularization: Regularization object to be applied.
        weights: The weights vector used for discrete integration.
            If none then Simpson's rule is used for computing the
            weights. If a callable object is passed, then the weight
            vector will be obtained by evaluating the object at the sample
            points of the passed FDataGrid object in the fit method. This
            parameter is only used when fitting a FDataGrid.
        components_basis: The basis in which we want the eigenvectors.
            We can use a different basis than the basis contained in the
            passed FDataBasis object. This parameter is only used when
            fitting a FDataBasis.
        alpha: Hyperparameter that controls the smoothness of the
            aproximation.
        eigenvectors: Eigenvectors of the covariance operator.
        eigenvalues: Eigenvalues of the covariance operator.

    Examples:
        >>> from skfda.misc.metrics import MahalanobisDistance
        >>> import numpy as np
        >>> from skfda.representation.grid import FDataGrid
        >>> data_matrix = np.array([[1.0, 0.0], [0.0, 2.0]])
        >>> grid_points = [0, 1]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>> mahalanobis = MahalanobisDistance(2)
        >>> mahalanobis.fit(fd)
        MahalanobisDistance(n_components=2)
        >>> mahalanobis(fd[0], fd[1])
        array([ 1.99680384])

    References:
        .. footbibliography::
    """

    def __init__(
        self,
        n_components: int = 10,
        centering: bool = True,
        regularization: Optional[TikhonovRegularization[FData]] = None,
        weights: Optional[Union[ArrayLike, WeightsCallable]] = None,
        components_basis: Optional[Basis] = None,
        alpha: float = 0.001,
        eigenvalues: Optional[NDArrayFloat] = None,
        eigenvectors: Optional[FData] = None,
    ) -> None:
        self.n_components = n_components
        self.centering = centering
        self.regularization = regularization
        self.weights = weights
        self.components_basis = components_basis
        self.alpha = alpha
        self.eigenvalues = eigenvalues
        self.eigenvectors = eigenvectors

    def fit(
        self,
        X: FData,
        y: None = None,
    ) -> MahalanobisDistance:
        """Fit the functional Mahalanobis distance to X.

        We extract the eigenvectors and corresponding eigenvalues
        of the covariance operator by using FPCA.

        Args:
            X: Stochastic process.
            y: Ignored.

        Returns:
            self
        """
        from ...preprocessing.dim_reduction import FPCA

        if self.eigenvalues is None or self.eigenvectors is None:
            fpca = FPCA(
                n_components=self.n_components,
                centering=self.centering,
                regularization=self.regularization,
                components_basis=self.components_basis,
                _weights=self.weights,
            )
            fpca.fit(X)
            self.eigenvalues_ = fpca.explained_variance_
            self.eigenvectors_ = fpca.components_

        else:
            self.eigenvalues_ = self.eigenvalues
            self.eigenvectors_ = self.eigenvectors

        return self

    def __call__(
        self,
        e1: FData,
        e2: FData,
    ) -> NDArrayFloat:
        """Compute the squared functional Mahalanobis distances of given observations.

        Args:
            e1: First object.
            e2: Second object.

        Returns:
            Squared functional Mahalanobis distance between two observations.
        """
        try:
            check_is_fitted(self)
            eigenvalues = self.eigenvalues_
            eigenvectors = self.eigenvectors_
        except NotFittedError:
            if self.eigenvalues is not None and self.eigenvectors is not None:
                eigenvalues = self.eigenvalues
                eigenvectors = self.eigenvectors
            else:
                raise

        return np.sum(  # type: ignore[no-any-return]
            eigenvalues
            * inner_product_matrix(e1 - e2, eigenvectors) ** 2
            / (eigenvalues + self.alpha)**2,
            axis=1,
        )



from __future__ import annotations

from abc import abstractmethod
from typing import Callable, Generic, TypeVar

from ...._utils._sklearn_adapter import BaseEstimator
from ....representation import FData
from ....typing._numpy import NDArrayFloat

Input = TypeVar("Input", bound=FData)


class CovarianceEstimator(
    BaseEstimator,
    Generic[Input],
):

    location_: Input
    covariance_: Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]

    def __init__(
        self,
        *,
        assume_centered: bool = False,
    ) -> None:
        self.assume_centered = assume_centered

    def _fit_mean(self, X: Input) -> None:
        self.location_ = X.mean()
        if self.assume_centered:
            self.location_ *= 0

    @abstractmethod
    def fit(self, X: Input, y: object = None) -> CovarianceEstimator[Input]:

        self._fit_mean(X)

        return self

    def score(self, X_test: Input, y: object = None) -> float:

        pass



"""Module with generic methods."""

from __future__ import annotations

import functools
import numbers
from functools import singledispatch
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Sized,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import scipy.integrate
from pandas.api.indexers import check_array_indexer
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import check_classification_targets
from typing_extensions import Literal, ParamSpec, Protocol

from ..typing._base import GridPoints, GridPointsLike
from ..typing._numpy import NDArrayAny, NDArrayFloat, NDArrayInt, NDArrayStr
from ._sklearn_adapter import BaseEstimator

ArrayDTypeT = TypeVar("ArrayDTypeT", bound="np.generic")

if TYPE_CHECKING:
    from ..representation import FData, FDataGrid
    from ..representation.basis import Basis
    from ..representation.extrapolation import ExtrapolationLike

    T = TypeVar("T", bound=FData)

    Input = TypeVar("Input", bound=Union[FData, NDArrayFloat])
    Output = TypeVar("Output", bound=Union[FData, NDArrayFloat])
    Target = TypeVar("Target", bound=NDArrayInt)


_MapAcceptableSelf = TypeVar(
    "_MapAcceptableSelf",
    bound="_MapAcceptable",
)


class _MapAcceptable(Protocol, Sized):

    def __getitem__(
        self: _MapAcceptableSelf,
        __key: Union[slice, NDArrayInt],  # noqa: WPS112
    ) -> _MapAcceptableSelf:
        pass

    @property
    def nbytes(self) -> int:
        pass


_MapAcceptableT = TypeVar(
    "_MapAcceptableT",
    bound=_MapAcceptable,
    contravariant=True,
)
MapFunctionT = TypeVar("MapFunctionT", covariant=True)
P = ParamSpec("P")


class _MapFunction(Protocol[_MapAcceptableT, P, MapFunctionT]):
    """Protocol for functions that can be mapped over several arrays."""

    def __call__(
        self,
        *args: _MapAcceptableT,
        **kwargs: P.kwargs,
    ) -> MapFunctionT:
        pass


class _PairwiseFunction(Protocol[_MapAcceptableT, P, MapFunctionT]):
    """Protocol for pairwise array functions."""

    def __call__(
        self,
        __arg1: _MapAcceptableT,  # noqa: WPS112
        __arg2: _MapAcceptableT,  # noqa: WPS112
        **kwargs: P.kwargs,  # type: ignore[name-defined]
    ) -> MapFunctionT:
        pass


def _to_grid(
    X: FData,
    y: FData,
    eval_points: Optional[NDArrayFloat] = None,
) -> Tuple[FDataGrid, FDataGrid]:
    """Transform a pair of FDatas in grids to perform calculations."""
    from .. import FDataGrid
    x_is_grid = isinstance(X, FDataGrid)
    y_is_grid = isinstance(y, FDataGrid)

    if eval_points is not None:
        X = X.to_grid(eval_points)
        y = y.to_grid(eval_points)
    elif x_is_grid and not y_is_grid:
        y = y.to_grid(X.grid_points[0])
    elif not x_is_grid and y_is_grid:
        X = X.to_grid(y.grid_points[0])
    elif not x_is_grid and not y_is_grid:
        X = X.to_grid()
        y = y.to_grid()

    return X, y


def _to_grid_points(grid_points_like: GridPointsLike) -> GridPoints:
    """Convert to grid points.

    If the original list is one-dimensional (e.g. [1, 2, 3]), return list to
    array (in this case [array([1, 2, 3])]).

    If the original list is two-dimensional (e.g. [[1, 2, 3], [4, 5]]), return
    a list containing other one-dimensional arrays (in this case
    [array([1, 2, 3]), array([4, 5])]).

    In any other case the behaviour is unespecified.

    """
    unidimensional = False

    if not isinstance(grid_points_like, Iterable):
        grid_points_like = [grid_points_like]

    if not isinstance(grid_points_like[0], Iterable):
        unidimensional = True

    if unidimensional:
        return (_int_to_real(np.asarray(grid_points_like)),)

    return tuple(_int_to_real(np.asarray(i)) for i in grid_points_like)


@overload
def _cartesian_product(
    axes: Sequence[np.typing.NDArray[ArrayDTypeT]],
    *,
    flatten: bool = True,
    return_shape: Literal[False] = False,
) -> np.typing.NDArray[ArrayDTypeT]:
    pass


@overload
def _cartesian_product(
    axes: Sequence[np.typing.NDArray[ArrayDTypeT]],
    *,
    flatten: bool = True,
    return_shape: Literal[True],
) -> Tuple[np.typing.NDArray[ArrayDTypeT], Tuple[int, ...]]:
    pass


def _cartesian_product(  # noqa: WPS234
    axes: Sequence[np.typing.NDArray[ArrayDTypeT]],
    *,
    flatten: bool = True,
    return_shape: bool = False,
) -> (
    np.typing.NDArray[ArrayDTypeT]
    | Tuple[np.typing.NDArray[ArrayDTypeT], Tuple[int, ...]]
):
    """
    Compute the cartesian product of the axes.

    Computes the cartesian product of the axes and returns a numpy array of
    1 dimension with all the possible combinations, for an arbitrary number of
    dimensions.

    Args:
        axes: List with axes.
        flatten: Whether to return the flatten array or keep one dimension per
            axis.
        return_shape: If ``True`` return the shape of the array before
            flattening.

    Returns:
        Numpy 2-D array with all the possible combinations.
        The entry (i,j) represent the j-th coordinate of the i-th point.
        If ``return_shape`` is ``True`` returns also the shape of the array
        before flattening.

    Examples:
        >>> from skfda._utils import _cartesian_product
        >>> axes = [[0,1],[2,3]]
        >>> _cartesian_product(axes)
        array([[0, 2],
               [0, 3],
               [1, 2],
               [1, 3]])

        >>> axes = [[0,1],[2,3],[4]]
        >>> _cartesian_product(axes)
        array([[0, 2, 4],
               [0, 3, 4],
               [1, 2, 4],
               [1, 3, 4]])

        >>> axes = [[0,1]]
        >>> _cartesian_product(axes)
        array([[0],
               [1]])
    """
    cartesian = np.stack(np.meshgrid(*axes, indexing='ij'), -1)

    shape = cartesian.shape

    if flatten:
        cartesian = cartesian.reshape(-1, len(axes))

    if return_shape:
        return cartesian, shape

    return cartesian  # type: ignore[no-any-return]


def _same_domain(fd: Union[Basis, FData], fd2: Union[Basis, FData]) -> bool:
    """Check if the domain range of two objects is the same."""
    return np.array_equal(fd.domain_range, fd2.domain_range)


def _one_grid_to_points(
    axes: GridPointsLike,
    *,
    dim_domain: int,
) -> Tuple[NDArrayFloat, Tuple[int, ...]]:
    """
    Convert a list of ndarrays, one per domain dimension, in the points.

    Returns also the shape containing the information of how each point
    is formed.
    """
    axes = _to_grid_points(axes)

    if len(axes) != dim_domain:
        raise ValueError(
            f"Length of axes should be {dim_domain}",
        )

    cartesian, shape = _cartesian_product(axes, return_shape=True)

    # Drop domain size dimension, as it is not needed to reshape the output
    shape = shape[:-1]

    return cartesian, shape


class EvaluateMethod(Protocol):
    """Evaluation method."""

    def __call__(
        self,
        __eval_points: NDArrayFloat,  # noqa: WPS112
        extrapolation: Optional[ExtrapolationLike],
        aligned: bool,
    ) -> NDArrayFloat:
        """Evaluate a function."""
        pass


@overload
def _evaluate_grid(
    axes: GridPointsLike,
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: Literal[True] = True,
) -> NDArrayFloat:
    pass


@overload
def _evaluate_grid(
    axes: Iterable[GridPointsLike],
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: Literal[False],
) -> NDArrayFloat:
    pass


@overload
def _evaluate_grid(
    axes: Union[GridPointsLike, Iterable[GridPointsLike]],
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: bool,
) -> NDArrayFloat:
    pass


def _evaluate_grid(  # noqa: WPS234
    axes: Union[GridPointsLike, Iterable[GridPointsLike]],
    *,
    evaluate_method: EvaluateMethod,
    n_samples: int,
    dim_domain: int,
    dim_codomain: int,
    extrapolation: Optional[ExtrapolationLike] = None,
    aligned: bool = True,
) -> NDArrayFloat:
    """
    Evaluate the functional object in the cartesian grid.

    This method is called internally by :meth:`evaluate` when the argument
    `grid` is True.

    Evaluates the functional object in the grid generated by the cartesian
    product of the axes. The length of the list of axes should be equal
    than the domain dimension of the object.

    If the list of axes has lengths :math:`n_1, n_2, ..., n_m`, where
    :math:`m` is equal than the dimension of the domain, the result of the
    evaluation in the grid will be a matrix with :math:`m+1` dimensions and
    shape :math:`n_{samples} x n_1 x n_2 x ... x n_m`.

    If `aligned` is false each sample is evaluated in a
    different grid, and the list of axes should contain a list of axes for
    each sample.

    If the domain dimension is 1, the result of the behaviour of the
    evaluation will be the same than :meth:`evaluate` without the grid
    option, but with worst performance.

    Args:
        axes: List of axes to generated the grid where the
            object will be evaluated.
        evaluate_method: Function used to evaluate the functional object.
        n_samples: Number of samples.
        dim_domain: Domain dimension.
        dim_codomain: Codomain dimension.
        extrapolation: Controls the
            extrapolation mode for elements outside the domain range. By
            default it is used the mode defined during the instance of the
            object.
        aligned: If False evaluates each sample
            in a different grid.

    Returns:
        Numpy array with dim_domain + 1 dimensions with
            the result of the evaluation.

    Raises:
        ValueError: If there are a different number of axes than the domain
            dimension.

    """

    # Compute intersection points and resulting shapes
    if aligned:

        axes = cast(GridPointsLike, axes)

        eval_points, shape = _one_grid_to_points(axes, dim_domain=dim_domain)

    else:

        axes_per_sample = cast(Iterable[GridPointsLike], axes)

        axes_per_sample = list(axes_per_sample)

        eval_points_tuple, shape_tuple = zip(
            *[
                _one_grid_to_points(a, dim_domain=dim_domain)
                for a in axes_per_sample
            ],
        )

        if len(eval_points_tuple) != n_samples:
            raise ValueError(
                "Should be provided a list of axis per sample",
            )

        eval_points = np.asarray(eval_points_tuple)

    # Evaluate the points
    evaluated = evaluate_method(
        eval_points,
        extrapolation=extrapolation,
        aligned=aligned,
    )

    # Reshape the result
    if aligned:

        res = evaluated.reshape(
            [n_samples] + list(shape) + [dim_codomain],
        )

    else:

        res = np.asarray([
            r.reshape(list(s) + [dim_codomain])
            for r, s in zip(evaluated, shape_tuple)
        ])

    return res


def nquad_vec(
    func: Callable[[NDArrayFloat], NDArrayFloat],
    ranges: Sequence[Tuple[float, float]],
) -> NDArrayFloat:
    """Perform multiple integration of vector valued functions."""
    initial_depth = len(ranges) - 1

    def integrate(*args: Any, depth: int) -> NDArrayFloat:  # noqa: WPS430

        if depth == 0:
            f = functools.partial(func, *args)
        else:
            f = functools.partial(integrate, *args, depth=depth - 1)

        return scipy.integrate.quad_vec(  # type: ignore[no-any-return]
            f,
            *ranges[initial_depth - depth],
        )[0]

    return integrate(depth=initial_depth)


def _map_in_batches(
    function: _MapFunction[_MapAcceptableT, P, np.typing.NDArray[ArrayDTypeT]],
    arguments: Tuple[_MapAcceptableT, ...],
    indexes: Tuple[NDArrayInt, ...],
    memory_per_batch: Optional[int] = None,
    *args: P.args,  # Should be empty
    **kwargs: P.kwargs,
) -> np.typing.NDArray[ArrayDTypeT]:
    """
    Map a function over samples of FData or ndarray tuples efficiently.

    This function prevents a large set of indexes to use all available
    memory and hang the PC.

    """
    if memory_per_batch is None:
        # 256MB is not too big
        memory_per_batch = 256 * 1024 * 1024  # noqa: WPS432

    memory_per_element = sum(a.nbytes // len(a) for a in arguments)
    n_elements_per_batch_allowed = memory_per_batch // memory_per_element
    if n_elements_per_batch_allowed < 1:
        raise ValueError("Too few memory allowed for the operation")

    n_indexes = len(indexes[0])

    assert all(n_indexes == len(i) for i in indexes)

    batches: List[np.typing.NDArray[ArrayDTypeT]] = []

    for pos in range(0, n_indexes, n_elements_per_batch_allowed):
        batch_args = tuple(
            a[i[pos:pos + n_elements_per_batch_allowed]]
            for a, i in zip(arguments, indexes)
        )

        batches.append(function(*batch_args, **kwargs))

    return np.concatenate(batches, axis=0)


def _pairwise_symmetric(
    function: _PairwiseFunction[
        _MapAcceptableT,
        P,
        np.typing.NDArray[ArrayDTypeT],
    ],
    arg1: _MapAcceptableT,
    arg2: Optional[_MapAcceptableT] = None,
    memory_per_batch: Optional[int] = None,
    *args: P.args,  # Should be empty
    **kwargs: P.kwargs,
) -> np.typing.NDArray[ArrayDTypeT]:
    """Compute pairwise a commutative function."""
    def map_function(
        *args: _MapAcceptableT,
        **kwargs: P.kwargs,
    ) -> np.typing.NDArray[ArrayDTypeT]:
        """Just to keep Mypy happy."""
        return function(args[0], args[1], **kwargs)

    dim1 = len(arg1)
    if arg2 is None or arg2 is arg1:
        triu_indices = np.triu_indices(dim1)

        triang_vec = _map_in_batches(
            map_function,
            (arg1, arg1),
            triu_indices,
            memory_per_batch,
            **kwargs,  # type: ignore[arg-type]
        )

        matrix = np.empty((dim1, dim1), dtype=triang_vec.dtype)

        # Set upper matrix
        matrix[triu_indices] = triang_vec

        # Set lower matrix
        matrix[(triu_indices[1], triu_indices[0])] = triang_vec

        return matrix

    dim2 = len(arg2)
    indices = np.indices((dim1, dim2))

    vec = _map_in_batches(
        map_function,
        (arg1, arg2),
        (indices[0].ravel(), indices[1].ravel()),
        memory_per_batch=memory_per_batch,
        **kwargs,  # type: ignore[arg-type]
    )

    return np.reshape(vec, (dim1, dim2))


def _int_to_real(array: Union[NDArrayInt, NDArrayFloat]) -> NDArrayFloat:
    """Convert integer arrays to floating point."""
    if np.issubdtype(array.dtype, np.integer):
        return array.astype(np.float64)

    assert np.issubdtype(array.dtype, np.floating)
    return cast(NDArrayFloat, array)


def _check_array_key(array: NDArrayAny, key: Any) -> Any:
    """Check a getitem key."""
    key = check_array_indexer(array, key)
    if isinstance(key, tuple):
        non_ellipsis = [i for i in key if i is not Ellipsis]
        if len(non_ellipsis) > 1:
            raise KeyError(key)
        key = non_ellipsis[0]

    if isinstance(key, numbers.Integral):  # To accept also numpy ints
        key = int(key)
        if key < 0:
            key = len(array) + key

        if not 0 <= key < len(array):
            raise IndexError("index out of bounds")

        return slice(key, key + 1)

    return key


def _check_estimator(estimator: Type[BaseEstimator]) -> None:
    from sklearn.utils.estimator_checks import (
        check_get_params_invariance,
        check_set_params,
    )

    name = estimator.__name__
    instance = estimator()
    check_get_params_invariance(name, instance)
    check_set_params(name, instance)


def _classifier_get_classes(
    y: NDArrayStr | NDArrayInt,
) -> Tuple[NDArrayStr | NDArrayInt, NDArrayInt]:

    check_classification_targets(y)

    le = LabelEncoder()
    y_ind = le.fit_transform(y)

    classes = le.classes_

    if classes.size < 2:
        raise ValueError(
            f'The number of classes has to be greater than'
            f'one; got {classes.size} class',
        )
    return classes, y_ind



"""Depth-based models for supervised classification."""
from __future__ import annotations

from collections import defaultdict
from contextlib import suppress
from itertools import combinations
from typing import (
    DefaultDict,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import pandas as pd
from scipy.interpolate import lagrange
from sklearn.base import clone
from sklearn.metrics import accuracy_score
from sklearn.pipeline import FeatureUnion, make_pipeline
from sklearn.utils.validation import check_is_fitted as sklearn_check_is_fitted

from ..._utils import _classifier_get_classes
from ..._utils._sklearn_adapter import BaseEstimator, ClassifierMixin
from ...exploratory.depth import Depth, ModifiedBandDepth
from ...preprocessing.feature_construction._per_class_transformer import (
    PerClassTransformer,
)
from ...representation import FData
from ...typing._numpy import NDArrayFloat, NDArrayInt, NDArrayStr

Input = TypeVar("Input", bound=FData)
Target = TypeVar("Target", bound=Union[NDArrayInt, NDArrayStr])


def _classifier_get_depth_methods(
    classes: NDArrayInt | NDArrayStr,
    X: Input,
    y_ind: NDArrayInt,
    depth_methods: Sequence[Depth[Input]],
) -> Sequence[Depth[Input]]:
    return [
        clone(depth_method).fit(X[y_ind == cur_class])
        for cur_class in range(len(classes))
        for depth_method in depth_methods
    ]


def _classifier_fit_depth_methods(
    X: Input,
    y: NDArrayInt | NDArrayStr,
    depth_methods: Sequence[Depth[Input]],
) -> Tuple[NDArrayStr | NDArrayInt, Sequence[Depth[Input]]]:
    classes, y_ind = _classifier_get_classes(y)

    class_depth_methods_ = _classifier_get_depth_methods(
        classes, X, y_ind, depth_methods,
    )

    return classes, class_depth_methods_


class DDClassifier(
    BaseEstimator,
    ClassifierMixin[Input, Target],
):
    """
    Depth-versus-depth (DD) classifer for functional data.

    Transforms the data into a DD-plot and then classifies using a polynomial
    of a chosen degree :footcite:p:`li++_2012_ddclassifier`.
    The polynomial passes through zero and maximizes the accuracy of the
    classification on the train dataset.

    If a point is below the polynomial in the DD-plot, it is classified to
    the first class. Otherwise, the point is classified to the second class.

    Parameters:
        degree: degree of the polynomial used to classify in the DD-plot
        depth_method:
            The depth class to use when calculating the depth of a test
            sample in a class. See the documentation of the depths module
            for a list of available depths. By default it is ModifiedBandDepth.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        We will fit a DD-classifier

        >>> from skfda.ml.classification import DDClassifier
        >>> clf = DDClassifier(degree=2)
        >>> clf.fit(X_train, y_train)
        DDClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test)
        array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDGClassifier`
        :class:`~skfda.ml.classification.MaximumDepthClassifier`
        :class:`~skfda.preprocessing.dim_reduction.feature_construction._ddg_transformer`

    References:
        .. footbibliography::
    """

    def __init__(
        self,
        degree: int,
        depth_method: Optional[Depth[Input]] = None,
    ) -> None:
        self.depth_method = depth_method
        self.degree = degree

    def fit(self, X: Input, y: Target) -> DDClassifier[Input, Target]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        if self.depth_method is None:
            self.depth_method = ModifiedBandDepth()

        classes, class_depth_methods = _classifier_fit_depth_methods(
            X, y, [self.depth_method],
        )

        self.classes_ = classes
        self.class_depth_methods_ = class_depth_methods

        if (len(self.classes_) != 2):
            raise ValueError("DDClassifier only accepts two classes.")

        dd_coordinates = [
            depth_method.transform(X)
            for depth_method in self.class_depth_methods_
        ]

        polynomial_elements = combinations(
            range(len(dd_coordinates[0])),  # noqa: WPS518
            self.degree,
        )

        accuracy = -1  # initialise accuracy

        for elements in polynomial_elements:
            x_coord = np.append(dd_coordinates[0][list(elements)], 0)
            y_coord = np.append(dd_coordinates[1][list(elements)], 0)

            poly = lagrange(
                x_coord, y_coord,
            )

            predicted_values = np.polyval(poly, dd_coordinates[0])

            y_pred = self.classes_[(
                dd_coordinates[1] > predicted_values
            ).astype(int)
            ]

            new_accuracy = accuracy_score(y, y_pred)

            if (new_accuracy > accuracy):
                accuracy = new_accuracy
                self.poly_ = poly

        return self

    def predict(self, X: Input) -> Target:
        """Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        sklearn_check_is_fitted(self)

        dd_coordinates = [
            depth_method.transform(X)
            for depth_method in self.class_depth_methods_
        ]

        predicted_values = np.polyval(self.poly_, dd_coordinates[0])

        return self.classes_[(  # type: ignore[no-any-return]
            dd_coordinates[1] > predicted_values
        ).astype(int)
        ]


class DDGClassifier(
    BaseEstimator,
    ClassifierMixin[Input, Target],
):
    r"""
    Generalized depth-versus-depth (DD) classifier for functional data.

    This classifier builds an interface around the DDGTransfomer.
    The transformer takes a list of k depths and performs the following map
    :footcite:p:`cuesta-albertos++_2017_ddgclassifier`:

    .. math::
        \mathcal{X} &\rightarrow \mathbb{R}^G \\
        x &\rightarrow \textbf{d} = (D_1^1(x), D_1^2(x),...,D_g^k(x))

    Where :math:`D_i^j(x)` is the depth of the point :math:`x` with respect to
    the data in the :math:`i`-th group using the :math:`j`-th depth of the
    provided list.
    Note that :math:`\mathcal{X}` is possibly multivariate, that is,
    :math:`\mathcal{X} = \mathcal{X}_1 \times ... \times \mathcal{X}_p`.

    In the G dimensional space the classification is performed using a
    multivariate classifer.

    Parameters:
        depth_method:
            The depth class or sequence of depths to use when calculating
            the depth of a test sample in a class. See the documentation of
            the depths module for a list of available depths. By default it
            is ModifiedBandDepth.
        multivariate_classifier:
            The multivariate classifier to use in the DDG-plot.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split

        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        >>> from skfda.exploratory.depth import (
        ...     ModifiedBandDepth,
        ...     IntegratedDepth,
        ... )
        >>> from sklearn.neighbors import KNeighborsClassifier

        We will fit a DDG-classifier using KNN

        >>> from skfda.ml.classification import DDGClassifier
        >>> clf = DDGClassifier(
        ...     depth_method=ModifiedBandDepth(),
        ...     multivariate_classifier=KNeighborsClassifier(),
        ... )
        >>> clf.fit(X_train, y_train)
        DDGClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test)
        array([1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

        It is also possible to use several depth functions to increase the
        number of features available to the classifier

        >>> clf = DDGClassifier(
        ...     depth_method=[
        ...         ("mbd", ModifiedBandDepth()),
        ...         ("id", IntegratedDepth()),
        ...     ],
        ...     multivariate_classifier=KNeighborsClassifier(),
        ... )
        >>> clf.fit(X_train, y_train)
        DDGClassifier(...)
        >>> clf.predict(X_test)
        array([1, 1, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])
        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDClassifier`
        :class:`~skfda.ml.classification.MaximumDepthClassifier`
        :class:`~skfda.preprocessing.dim_reduction.feature_construction._ddg_transformer`

    References:
        .. footbibliography::

    """

    def __init__(  # noqa: WPS234
        self,
        *,
        multivariate_classifier: ClassifierMixin[
            NDArrayFloat,
            NDArrayInt,
        ] | None = None,
        depth_method: Depth[Input] | Sequence[
            Tuple[str, Depth[Input]]
        ] | None = None,
    ) -> None:
        self.multivariate_classifier = multivariate_classifier
        self.depth_method = depth_method

    def get_params(self, deep: bool = True) -> Mapping[str, object]:
        params = BaseEstimator.get_params(self, deep=deep)
        if deep and isinstance(self.depth_method, Sequence):
            for name, depth in self.depth_method:
                depth_params = depth.get_params(deep=deep)

                for key, value in depth_params.items():
                    params[f"depth_method__{name}__{key}"] = value

        return params  # type: ignore[no-any-return]

    # Copied from scikit-learn's _BaseComposition with minor modifications
    def _set_params(
        self,
        attr: str,
        **params: object,
    ) -> DDGClassifier[Input, Target]:
        # Ensure strict ordering of parameter setting:
        # 1. All steps
        if attr in params:
            setattr(self, attr, params.pop(attr))
        # 2. Replace items with estimators in params
        items = getattr(self, attr)
        if isinstance(items, list) and items:
            # Get item names used to identify valid names in params
            # `zip` raises a TypeError when `items` does not contains
            # elements of length 2
            with suppress(TypeError):
                item_names, _ = zip(*items)
                item_params: DefaultDict[
                    str,
                    Dict[str, object],
                ] = defaultdict(dict)
                for name in list(params.keys()):
                    if name.startswith(f"{attr}__"):
                        key, delim, sub_key = name.partition("__")
                        if "__" not in sub_key and sub_key in item_names:
                            self._replace_estimator(
                                attr,
                                sub_key,
                                params.pop(name),
                            )
                        else:
                            key, delim, sub_key = sub_key.partition("__")
                            item_params[key][sub_key] = params.pop(name)

                for name, estimator in items:
                    estimator.set_params(**item_params[name])

        # 3. Step parameters and other initialisation arguments
        super().set_params(**params)
        return self

    # Copied from scikit-learn's _BaseComposition
    def _replace_estimator(
        self,
        attr: str,
        name: str,
        new_val: object,
    ) -> None:
        # assumes `name` is a valid estimator name
        new_estimators = list(getattr(self, attr))
        for i, (estimator_name, _) in enumerate(new_estimators):
            if estimator_name == name:
                new_estimators[i] = (name, new_val)
                break
        setattr(self, attr, new_estimators)

    def set_params(
        self,
        **params: object,
    ) -> DDGClassifier[Input, Target]:

        return self._set_params("depth_method", **params)

    def fit(self, X: Input, y: Target) -> DDGClassifier[Input, Target]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: FDataGrid with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        depth_method = (
            ModifiedBandDepth()
            if self.depth_method is None
            else self.depth_method
        )

        if isinstance(depth_method, Sequence):
            transformer = FeatureUnion([
                (name, PerClassTransformer(depth))
                for name, depth in depth_method
            ])
        else:
            transformer = PerClassTransformer(depth_method)

        self._pipeline = make_pipeline(
            transformer,
            clone(self.multivariate_classifier),
        )

        self._pipeline.fit(X, y)
        self.classes_ = _classifier_get_classes(y)[0]

        return self

    def predict(self, X: Input) -> Target:
        """Predict the class labels for the provided data.

        Args:
            X: FDataGrid with the test samples.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        return self._pipeline.predict(X)  # type: ignore[no-any-return]


class _ArgMaxClassifier(
    BaseEstimator,
    ClassifierMixin[NDArrayFloat, Target],
):
    r"""Arg max classifier for multivariate data.

    Test samples are classified to the class that corresponds to the
    index of the highest coordinate.

    Examples:
        >>> import numpy as np
        >>> X = np.array([[1,5], [3,2], [4,1]])
        >>> y = np.array([1, 0, 0])

        We will fit an ArgMax classifier

        >>> from skfda.ml.classification._depth_classifiers import \
        ... _ArgMaxClassifier
        >>> clf = _ArgMaxClassifier()
        >>> clf.fit(X, y)
        _ArgMaxClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X) # Predict labels for test samples
        array([1, 0, 0])
    """

    def fit(self, X: NDArrayFloat, y: Target) -> _ArgMaxClassifier[Target]:
        """Fit the model using X as training data and y as target values.

        Args:
            X: Array with the training data.
            y: Target values of shape = (n_samples).

        Returns:
            self
        """
        classes, _ = _classifier_get_classes(y)
        self.classes_ = classes
        return self

    def predict(self, X: Union[NDArrayFloat, pd.DataFrame]) -> Target:
        """Predict the class labels for the provided data.

        Args:
            X: Array with the test samples or a pandas DataFrame.

        Returns:
            Array of shape (n_samples) with class labels
                for each data sample.
        """
        if isinstance(X, pd.DataFrame):
            X = X.to_numpy()
        return self.classes_[  # type: ignore[no-any-return]
            np.argmax(X, axis=1)
        ]


class MaximumDepthClassifier(DDGClassifier[Input, Target]):
    """
    Maximum depth classifier for functional data.

    Test samples are classified to the class where they are deeper
    :footcite:p:`ghosh+chaudhuri_2005_maximum`.

    Parameters:
        depth_method:
            The depth class to use when calculating the depth of a test
            sample in a class. See the documentation of the depths module
            for a list of available depths. By default it is ModifiedBandDepth.

    Examples:
        Firstly, we will import and split the Berkeley Growth Study dataset

        >>> from skfda.datasets import fetch_growth
        >>> from sklearn.model_selection import train_test_split
        >>> dataset = fetch_growth()
        >>> fd = dataset['data']
        >>> y = dataset['target']
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     fd, y, test_size=0.25, stratify=y, random_state=0)

        We will fit a Maximum depth classifier

        >>> from skfda.ml.classification import MaximumDepthClassifier
        >>> clf = MaximumDepthClassifier()
        >>> clf.fit(X_train, y_train)
        MaximumDepthClassifier(...)

        We can predict the class of new samples

        >>> clf.predict(X_test) # Predict labels for test samples
        array([1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1,
               1, 0, 0, 0, 0, 1, 1, 1, 1, 0, 1, 1])

        Finally, we calculate the mean accuracy for the test data

        >>> clf.score(X_test, y_test)
        0.875

    See also:
        :class:`~skfda.ml.classification.DDClassifier`
        :class:`~skfda.ml.classification.DDGClassifier`

    References:
        .. footbibliography::

    """

    def __init__(self, depth_method: Depth[Input] | None = None) -> None:
        super().__init__(
            multivariate_classifier=_ArgMaxClassifier(),
            depth_method=depth_method,
        )



"""Tests for Hotelling functions."""
import unittest

from skfda.inference.hotelling import hotelling_t2, hotelling_test_ind
from skfda.representation import FDataGrid
from skfda.representation.basis import FourierBasis


class HotellingTests(unittest.TestCase):
    """Tests for Hotelling statistic and test."""

    def test_hotelling_test_ind_args(self) -> None:
        """Test that invalid arguments are rejected in test."""
        fd1 = FDataGrid([[1, 1, 1]])

        with self.assertRaises(TypeError):
            hotelling_test_ind(fd1.to_basis(FourierBasis(n_basis=3)), fd1)
        with self.assertRaises(TypeError):
            hotelling_test_ind(fd1, fd1.to_basis(FourierBasis(n_basis=3)))
        with self.assertRaises(ValueError):
            hotelling_test_ind(fd1, fd1, n_reps=0)

    def test_hotelling_t2_args(self) -> None:
        """Test that invalid arguments are rejected in statistic."""
        fd1 = FDataGrid([[1, 1, 1]])

        with self.assertRaises(TypeError):
            hotelling_t2(fd1.to_basis(FourierBasis(n_basis=3)), fd1)
        with self.assertRaises(TypeError):
            hotelling_t2(fd1, fd1.to_basis(FourierBasis(n_basis=3)))

    def test_hotelling_t2(self) -> None:
        """Trivial checks for the statistic."""
        fd1 = FDataGrid([[1, 1, 1], [1, 1, 1]])
        fd2 = FDataGrid([[1, 1, 1], [2, 2, 2]])
        self.assertAlmostEqual(hotelling_t2(fd1, fd1), 0)
        self.assertAlmostEqual(hotelling_t2(fd1, fd2), 1)

        fd1 = fd1.to_basis(FourierBasis(n_basis=3))
        fd2 = fd2.to_basis(FourierBasis(n_basis=3))
        self.assertAlmostEqual(hotelling_t2(fd1, fd1), 0)
        self.assertAlmostEqual(hotelling_t2(fd1, fd2), 1)

    def test_hotelling_test(self) -> None:
        """Trivial checks for the test."""
        fd1 = FDataGrid([[1, 1, 1], [1, 1, 1]])
        fd2 = FDataGrid([[3, 3, 3], [2, 2, 2]])
        t2, pval, dist = hotelling_test_ind(
            fd1,
            fd2,
            return_dist=True,
            random_state=0,
        )
        self.assertAlmostEqual(t2, 9)
        self.assertAlmostEqual(pval, 0)
        self.assertEqual(len(dist), 6)
        reps = 5
        t2, pval, dist = hotelling_test_ind(
            fd1,
            fd2,
            return_dist=True,
            n_reps=reps,
            random_state=1,
        )
        self.assertEqual(len(dist), reps)


if __name__ == '__main__':
    unittest.main()



from functools import singledispatch
from typing import Any, Callable, TypeVar

import numpy as np

from ....representation import FDataGrid
from ....typing._numpy import NDArrayAny, NDArrayFloat

dtype_bound = np.number
X_T = TypeVar("X_T")
dtype_X_T = TypeVar("dtype_X_T", bound="dtype_bound[Any]")
dtype_y_T = TypeVar("dtype_y_T", bound="dtype_bound[Any]")

depX_T = TypeVar("depX_T", bound=NDArrayAny)
depy_T = TypeVar("depy_T", bound=NDArrayAny)

_DependenceMeasure = Callable[
    [depX_T, depy_T],
    NDArrayFloat,
]


@singledispatch
def _compute_dependence(
    X: X_T,
    y: np.typing.NDArray[dtype_y_T],
    *,
    dependence_measure: _DependenceMeasure[
        np.typing.NDArray[dtype_X_T],
        np.typing.NDArray[dtype_y_T],
    ],
) -> X_T:
    """
    Compute dependence between points and target.

    Computes the dependence of each point in each trajectory in X with the
    corresponding class label in Y.

    """
    from dcor import rowwise

    assert isinstance(X, np.ndarray)
    X_ndarray = X

    # Shape without number of samples and codomain dimension
    input_shape = X_ndarray.shape[1:-1]

    # Move n_samples to the end
    # The shape is now input_shape + n_samples + n_output
    X_ndarray = np.moveaxis(X_ndarray, 0, -2)

    # Join input in a list for rowwise
    X_ndarray = X_ndarray.reshape(-1, X_ndarray.shape[-2], X_ndarray.shape[-1])

    if y.ndim == 1:
        y = np.atleast_2d(y).T

    Y = np.array([y] * len(X_ndarray))

    dependence_results = rowwise(  # type: ignore[no-untyped-call]
        dependence_measure,
        X_ndarray,
        Y,
    )

    return dependence_results.reshape(  # type: ignore[no-any-return]
        input_shape,
    )


@_compute_dependence.register
def _compute_dependence_fdatagrid(
    X: FDataGrid,
    y: np.typing.NDArray[dtype_y_T],
    *,
    dependence_measure: _DependenceMeasure[
        np.typing.NDArray[dtype_X_T],
        np.typing.NDArray[dtype_y_T],
    ],
) -> FDataGrid:

    return X.copy(
        data_matrix=_compute_dependence(
            X.data_matrix,
            y,
            dependence_measure=dependence_measure,
        ),
        coordinate_names=("relevance",),
        sample_names=("relevance function",),
    )
